# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""
Rooney Biegler model, based on Rooney, W. C. and Biegler, L. T. (2001). Design for
model parameter uncertainty using nonlinear confidence regions. AIChE Journal,
47(8), 1794-1804.
"""

from pyomo.common.dependencies import pandas as pd
import pyomo.environ as pyo
from pyomo.contrib.parmest.experiment import Experiment


def rooney_biegler_model(data, theta=None):
    """Create a Pyomo model for the Rooney-Biegler parameter estimation problem.

    This model implements an exponential response function based on Rooney & Biegler (2001).
    The response is: y = asymptote * (1 - exp(-rate_constant * hour))

    Parameters
    ----------
    data : pandas.DataFrame
        DataFrame containing 'hour' and 'y' columns for the experiment data point.
    theta : dict, optional
        Dictionary with 'asymptote' and 'rate_constant' keys for parameter initialization.
        Default is {'asymptote': 15, 'rate_constant': 0.5}.

    Returns
    -------
    pyo.ConcreteModel
        A Pyomo ConcreteModel with fixed parameters, experiment inputs (hour),
        and response variable (y) with the exponential constraint.
    """
    model = pyo.ConcreteModel()

    if theta is None:
        theta = {'asymptote': 15, 'rate_constant': 0.5}

    model.asymptote = pyo.Var(initialize=theta['asymptote'])
    model.rate_constant = pyo.Var(initialize=theta['rate_constant'])

    # Fix the unknown parameters
    model.asymptote.fix()
    model.rate_constant.fix()

    # Add the experiment inputs
    model.hour = pyo.Var(initialize=data["hour"].iloc[0], bounds=(0, 10))

    # Fix the experiment inputs
    model.hour.fix()

    # Add the response variable
    model.y = pyo.Var(within=pyo.PositiveReals, initialize=data["y"].iloc[0])

    def response_rule(m):
        return m.y == m.asymptote * (1 - pyo.exp(-m.rate_constant * m.hour))

    model.response_function = pyo.Constraint(rule=response_rule)

    return model


class RooneyBieglerExperiment(Experiment):
    """Experiment class for Rooney-Biegler parameter estimation and design of experiments.

    This class wraps the Rooney-Biegler exponential model for use with Pyomo's
    parmest (parameter estimation) and DoE (Design of Experiments) tools.
    """

    def __init__(self, data, measure_error=None, theta=None):
        """Initialize a Rooney-Biegler experiment instance.

        Parameters
        ----------
        data : pandas.Series or dict
            Experiment data containing 'hour' (time input) and 'y' (response) values.
        measure_error : float, optional
            Standard deviation of measurement error for the response variable.
            Required for DoE and covariance estimation. Default is None.
        theta : dict, optional
            Initial parameter values with 'asymptote' and 'rate_constant' keys.
            Default is None, which uses {'asymptote': 15, 'rate_constant': 0.5}.
        """
        self.data = data
        self.model = None
        self.measure_error = measure_error
        self.theta = theta

    def create_model(self):
        # rooney_biegler_model expects a dataframe
        if hasattr(self.data, 'to_frame'):
            # self.data is a pandas Series
            data_df = self.data.to_frame().transpose()
        else:
            # self.data is a dict
            data_df = pd.DataFrame([self.data])
        self.model = rooney_biegler_model(data_df, theta=self.theta)

    def label_model(self):

        m = self.model

        # Add experiment outputs as a suffix
        # Experiment outputs suffix is required for parmest
        m.experiment_outputs = pyo.Suffix(direction=pyo.Suffix.LOCAL)
        m.experiment_outputs.update([(m.y, self.data['y'])])

        # Add unknown parameters as a suffix
        # Unknown parameters suffix is required for both Pyomo.DoE and parmest
        m.unknown_parameters = pyo.Suffix(direction=pyo.Suffix.LOCAL)
        m.unknown_parameters.update(
            (k, pyo.value(k)) for k in [m.asymptote, m.rate_constant]
        )

        # Add measurement error as a suffix
        # Measurement error suffix is required for Pyomo.DoE and
        #  `cov` estimation in parmest
        m.measurement_error = pyo.Suffix(direction=pyo.Suffix.LOCAL)
        m.measurement_error.update([(m.y, self.measure_error)])

        # Add hour as an experiment input
        # Experiment inputs suffix is required for Pyomo.DoE
        m.experiment_inputs = pyo.Suffix(direction=pyo.Suffix.LOCAL)
        m.experiment_inputs.update([(m.hour, self.data['hour'])])

    def get_labeled_model(self):
        if self.model is None:
            self.create_model()
            self.label_model()
        return self.model



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.collections import ComponentSet
from pyomo.common.numeric_types import native_numeric_types
from pyomo.core import Var
from pyomo.core.expr.logical_expr import _flattened
from pyomo.core.expr.numeric_expr import (
    LinearExpression,
    MonomialTermExpression,
    ProductExpression,
    SumExpression,
)
from pyomo.repn.util import (
    ExitNodeDispatcher,
    ExprType,
    initialize_exit_node_dispatcher,
    check_constant,
)
import pyomo.repn.linear as linear
import pyomo.repn.quadratic as quadratic

_FIXED = ExprType.FIXED
_CONSTANT = ExprType.CONSTANT
_LINEAR = ExprType.LINEAR
_GENERAL = ExprType.GENERAL


class ParameterizedRepnMixin:
    @staticmethod
    def constant_flag(val):
        if val.__class__ in native_numeric_types:
            return val
        return 2  # something not 0 or 1

    @staticmethod
    def multiplier_flag(val):
        if val.__class__ in native_numeric_types:
            if not val:
                return 2
            return val
        return 2  # something not 0 or 1

    def walker_exitNode(self):
        ans = super().walker_exitNode()
        if ans[0] is _CONSTANT and (
            self.constant.__class__ not in native_numeric_types
            or self.multiplier.__class__ not in native_numeric_types
        ):
            return _FIXED, ans[1]
        return ans


class ParameterizedBeforeChildDispatcher(linear.LinearBeforeChildDispatcher):
    def __init__(self):
        super().__init__()
        self[Var] = self._before_var
        self[MonomialTermExpression] = self._before_monomial
        self[LinearExpression] = self._before_linear
        self[SumExpression] = self._before_general_expression

    @staticmethod
    def _before_linear(visitor, child):
        return True, None

    @staticmethod
    def _before_monomial(visitor, child):
        return True, None

    @staticmethod
    def _before_general_expression(visitor, child):
        return True, None

    @staticmethod
    def _before_var(visitor, child):
        _id = id(child)
        if _id not in visitor.var_map:
            if child in visitor.wrt:
                # pseudo-constant, and we need to leave it in even if it is fixed!
                # We aren't treating this Var as a Var for the purposes of this walker
                return False, (_FIXED, child)
            if child.fixed:
                return False, (_CONSTANT, check_constant(child.value, child, visitor))
            # This is a normal situation
            visitor.var_recorder.add(child)
        ans = visitor.Result()
        ans.linear[_id] = 1
        return False, (ExprType.LINEAR, ans)


def _handle_product_constant_constant(visitor, node, arg1, arg2):
    # [ESJ 5/22/24]: Overriding this handler to exclude the deprecation path for
    # 0 * nan. It doesn't need overridden when that deprecation path goes away.
    return _CONSTANT, arg1[1] * arg2[1]


def update_exit_node_handlers(exit_node_handlers=None):
    exit_node_handlers[ProductExpression][
        (_CONSTANT, _CONSTANT)
    ] = _handle_product_constant_constant
    return exit_node_handlers


class ParameterizedRepnVisitorMixin:
    def __init__(
        self,
        subexpression_cache,
        var_map=None,
        var_order=None,
        sorter=None,
        wrt=None,
        var_recorder=None,
    ):
        super().__init__(
            subexpression_cache=subexpression_cache,
            var_map=var_map,
            var_order=var_order,
            sorter=sorter,
            var_recorder=var_recorder,
        )
        if wrt is None:
            raise ValueError(f"{self.__class__.__name__}: wrt not specified")
        self.wrt = ComponentSet(_flattened(wrt))

    def finalizeResult(self, result):
        ans = result[1]
        if ans.__class__ is self.Result and not ans.constant_flag(ans.multiplier):
            self._factor_multiplier_into_ans(ans, 0)
            return ans
        return super().finalizeResult(result)


class ParameterizedLinearRepn(ParameterizedRepnMixin, linear.LinearRepn):
    pass


class ParameterizedLinearRepnVisitor(
    ParameterizedRepnVisitorMixin, linear.LinearRepnVisitor
):
    Result = ParameterizedLinearRepn
    before_child_dispatcher = ParameterizedBeforeChildDispatcher()
    exit_node_dispatcher = ExitNodeDispatcher(
        linear.initialize_exit_node_dispatcher(
            update_exit_node_handlers(linear.define_exit_node_handlers())
        )
    )


class ParameterizedQuadraticRepn(ParameterizedRepnMixin, quadratic.QuadraticRepn):
    pass


class ParameterizedQuadraticRepnVisitor(
    ParameterizedRepnVisitorMixin, quadratic.QuadraticRepnVisitor
):
    Result = ParameterizedQuadraticRepn
    before_child_dispatcher = ParameterizedBeforeChildDispatcher()
    exit_node_dispatcher = ExitNodeDispatcher(
        initialize_exit_node_dispatcher(
            update_exit_node_handlers(quadratic.define_exit_node_handlers())
        )
    )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pickle
import math

import pyomo.common.unittest as unittest
from pyomo.kernel import pprint, IntegerSet
from pyomo.core.kernel.base import ICategorizedObject
from pyomo.core.kernel.constraint import (
    IConstraint,
    linear_constraint,
    constraint,
    constraint_dict,
    constraint_tuple,
    constraint_list,
)
from pyomo.core.kernel.variable import variable, variable_tuple
from pyomo.core.kernel.block import block
from pyomo.core.kernel.parameter import parameter
from pyomo.core.kernel.expression import expression, data_expression
from pyomo.core.kernel.conic import (
    _build_linking_constraints,
    quadratic,
    rotated_quadratic,
    primal_exponential,
    primal_power,
    dual_exponential,
    dual_power,
    primal_geomean,
    dual_geomean,
)


class _conic_tester_base:
    _object_factory = None

    def setUp(self):
        assert self._object_factory is not None

    def test_pprint(self):
        # Not really testing what the output is, just that
        # an error does not occur. The pprint functionality
        # is still in the early stages.
        c = self._object_factory()
        pprint(c)
        b = block()
        b.c = c
        pprint(c)
        pprint(b)
        m = block()
        m.b = b
        pprint(c)
        pprint(b)
        pprint(m)

    def test_type(self):
        c = self._object_factory()
        self.assertTrue(isinstance(c, ICategorizedObject))
        self.assertTrue(isinstance(c, IConstraint))

    def test_ctype(self):
        c = self._object_factory()
        self.assertIs(c.ctype, IConstraint)
        self.assertIs(type(c)._ctype, IConstraint)

    def test_pickle(self):
        c = self._object_factory()
        self.assertIs(c.lb, None)
        self.assertEqual(c.ub, 0)
        self.assertIsNot(c.body, None)
        self.assertIs(c.parent, None)
        cup = pickle.loads(pickle.dumps(c))
        self.assertIs(cup.lb, None)
        self.assertEqual(cup.ub, 0)
        self.assertIsNot(cup.body, None)
        self.assertIs(cup.parent, None)
        b = block()
        b.c = c
        self.assertIs(c.parent, b)
        bup = pickle.loads(pickle.dumps(b))
        cup = bup.c
        self.assertIs(cup.lb, None)
        self.assertEqual(cup.ub, 0)
        self.assertIsNot(cup.body, None)
        self.assertIs(cup.parent, bup)

    def test_properties(self):
        c = self._object_factory()
        self.assertIs(c._body, None)
        self.assertIs(c.parent, None)
        self.assertEqual(c.has_lb(), False)
        self.assertIs(c.lb, None)
        self.assertEqual(c.has_ub(), True)
        self.assertEqual(c.ub, 0)
        self.assertEqual(c.equality, False)
        self.assertEqual(c.check_convexity_conditions(), True)
        self.assertEqual(c.check_convexity_conditions(relax=False), True)
        self.assertEqual(c.check_convexity_conditions(relax=True), True)

        with self.assertRaises(AttributeError):
            c.lb = 1
        with self.assertRaises(AttributeError):
            c.ub = 1
        with self.assertRaises(AttributeError):
            c.rhs = 1
        with self.assertRaises(ValueError):
            c.rhs
        with self.assertRaises(AttributeError):
            c.equality = True

        self.assertIs(c._body, None)
        self.assertIs(c.parent, None)
        self.assertEqual(c.has_lb(), False)
        self.assertIs(c.lb, None)
        self.assertEqual(c.has_ub(), True)
        self.assertEqual(c.ub, 0)
        self.assertEqual(c.equality, False)
        self.assertEqual(c.check_convexity_conditions(), True)
        self.assertEqual(c.check_convexity_conditions(relax=False), True)
        self.assertEqual(c.check_convexity_conditions(relax=True), True)

        self.assertEqual(c.active, True)
        with self.assertRaises(AttributeError):
            c.active = False
        self.assertEqual(c.active, True)
        c.deactivate()
        self.assertEqual(c.active, False)
        c.activate()
        self.assertEqual(c.active, True)

    def test_containers(self):
        c = self._object_factory()
        self.assertIs(c.parent, None)
        cdict = constraint_dict()
        cdict[None] = c
        self.assertIs(c.parent, cdict)
        del cdict[None]
        self.assertIs(c.parent, None)
        clist = constraint_list()
        clist.append(c)
        self.assertIs(c.parent, clist)
        clist.remove(c)
        self.assertIs(c.parent, None)
        ctuple = constraint_tuple((c,))
        self.assertIs(c.parent, ctuple)


class Test_quadratic(_conic_tester_base, unittest.TestCase):
    _object_factory = lambda self: quadratic(
        r=variable(lb=0), x=[variable(), variable()]
    )

    def test_expression(self):
        c = self._object_factory()
        self.assertIs(c._body, None)
        with self.assertRaises(ValueError):
            self.assertIs(c(), None)
        with self.assertRaises(ValueError):
            self.assertIs(c(exception=True), None)
        self.assertIs(c(exception=False), None)
        self.assertIs(c._body, None)
        self.assertIs(c.slack, None)
        self.assertIs(c.lslack, None)
        self.assertIs(c.uslack, None)

        c.r.value = 3
        c.x[0].value = 5
        c.x[1].value = 2
        val = 5**2 + 2**2 - 3**2
        self.assertEqual(c(), val)
        self.assertEqual(c.slack, -val)
        self.assertEqual(c.lslack, float('inf'))
        self.assertEqual(c.uslack, -val)
        self.assertIs(c._body, None)
        # check body
        self.assertEqual(c.body(), val)
        self.assertEqual(c(), val)
        self.assertEqual(c.slack, -val)
        self.assertEqual(c.lslack, float('inf'))
        self.assertEqual(c.uslack, -val)
        self.assertIsNot(c._body, None)

    def test_check_convexity_conditions(self):
        c = self._object_factory()
        self.assertEqual(c.check_convexity_conditions(), True)

        c = self._object_factory()
        c.r.domain_type = IntegerSet
        self.assertEqual(c.check_convexity_conditions(), False)
        self.assertEqual(c.check_convexity_conditions(relax=True), True)
        c = self._object_factory()
        c.r.lb = None
        self.assertEqual(c.check_convexity_conditions(), False)
        c = self._object_factory()
        c.r.lb = -1
        self.assertEqual(c.check_convexity_conditions(), False)

        c = self._object_factory()
        c.x[0].domain_type = IntegerSet
        self.assertEqual(c.check_convexity_conditions(), False)
        self.assertEqual(c.check_convexity_conditions(relax=True), True)

    def test_as_domain(self):
        ret = quadratic.as_domain(r=3, x=[1, 2])
        self.assertIs(type(ret), block)
        q, c, r, x = ret.q, ret.c, ret.r, ret.x
        self.assertEqual(q.check_convexity_conditions(), True)
        self.assertIs(type(q), quadratic)
        self.assertIs(type(x), variable_tuple)
        self.assertIs(type(r), variable)
        self.assertEqual(len(x), 2)
        self.assertIs(type(c), constraint_tuple)
        self.assertEqual(len(c), 3)
        self.assertEqual(c[0].rhs, 3)
        r.value = 3
        self.assertEqual(c[0].slack, 0)
        r.value = None
        self.assertEqual(c[1].rhs, 1)
        x[0].value = 1
        self.assertEqual(c[1].slack, 0)
        x[0].value = None
        self.assertEqual(c[2].rhs, 2)
        x[1].value = 2
        self.assertEqual(c[2].slack, 0)
        x[1].value = None


class Test_rotated_quadratic(_conic_tester_base, unittest.TestCase):
    _object_factory = lambda self: rotated_quadratic(
        r1=variable(lb=0), r2=variable(lb=0), x=[variable(), variable()]
    )

    def test_expression(self):
        c = self._object_factory()
        self.assertIs(c._body, None)
        with self.assertRaises(ValueError):
            self.assertIs(c(), None)
        with self.assertRaises(ValueError):
            self.assertIs(c(exception=True), None)
        self.assertIs(c(exception=False), None)
        self.assertIs(c._body, None)
        self.assertIs(c.slack, None)
        self.assertIs(c.lslack, None)
        self.assertIs(c.uslack, None)

        c.r1.value = 5
        c.r2.value = 7
        c.x[0].value = 2
        c.x[1].value = 3
        val = 2**2 + 3**2 - 2 * 5 * 7
        self.assertEqual(c(), val)
        self.assertEqual(c.slack, -val)
        self.assertEqual(c.lslack, float('inf'))
        self.assertEqual(c.uslack, -val)
        self.assertIs(c._body, None)
        # check body
        self.assertEqual(c.body(), val)
        self.assertEqual(c(), val)
        self.assertEqual(c.slack, -val)
        self.assertEqual(c.lslack, float('inf'))
        self.assertEqual(c.uslack, -val)
        self.assertIsNot(c._body, None)

    def test_check_convexity_conditions(self):
        c = self._object_factory()
        self.assertEqual(c.check_convexity_conditions(), True)

        c = self._object_factory()
        c.r1.domain_type = IntegerSet
        self.assertEqual(c.check_convexity_conditions(), False)
        self.assertEqual(c.check_convexity_conditions(relax=True), True)
        c = self._object_factory()
        c.r1.lb = None
        self.assertEqual(c.check_convexity_conditions(), False)
        c = self._object_factory()
        c.r1.lb = -1
        self.assertEqual(c.check_convexity_conditions(), False)

        c = self._object_factory()
        c.r2.domain_type = IntegerSet
        self.assertEqual(c.check_convexity_conditions(), False)
        self.assertEqual(c.check_convexity_conditions(relax=True), True)
        c = self._object_factory()
        c.r2.lb = None
        self.assertEqual(c.check_convexity_conditions(), False)
        c = self._object_factory()
        c.r2.lb = -1
        self.assertEqual(c.check_convexity_conditions(), False)

        c = self._object_factory()
        c.x[0].domain_type = IntegerSet
        self.assertEqual(c.check_convexity_conditions(), False)
        self.assertEqual(c.check_convexity_conditions(relax=True), True)

    def test_as_domain(self):
        ret = rotated_quadratic.as_domain(r1=3, r2=4, x=[1, 2])
        self.assertIs(type(ret), block)
        q, c, r1, r2, x = ret.q, ret.c, ret.r1, ret.r2, ret.x
        self.assertEqual(q.check_convexity_conditions(), True)
        self.assertIs(type(q), rotated_quadratic)
        self.assertIs(type(x), variable_tuple)
        self.assertIs(type(r1), variable)
        self.assertIs(type(r2), variable)
        self.assertEqual(len(x), 2)
        self.assertIs(type(c), constraint_tuple)
        self.assertEqual(len(c), 4)
        self.assertEqual(c[0].rhs, 3)
        r1.value = 3
        self.assertEqual(c[0].slack, 0)
        r1.value = None
        self.assertEqual(c[1].rhs, 4)
        r2.value = 4
        self.assertEqual(c[1].slack, 0)
        r2.value = None
        self.assertEqual(c[2].rhs, 1)
        x[0].value = 1
        self.assertEqual(c[2].slack, 0)
        x[0].value = None
        self.assertEqual(c[3].rhs, 2)
        x[1].value = 2
        self.assertEqual(c[3].slack, 0)
        x[1].value = None


class Test_primal_exponential(_conic_tester_base, unittest.TestCase):
    _object_factory = lambda self: primal_exponential(
        r=variable(lb=0), x1=variable(lb=0), x2=variable()
    )

    def test_expression(self):
        c = self._object_factory()
        self.assertIs(c._body, None)
        with self.assertRaises(ValueError):
            self.assertIs(c(), None)
        with self.assertRaises(ValueError):
            self.assertIs(c(exception=True), None)
        self.assertIs(c(exception=False), None)
        self.assertIs(c._body, None)
        self.assertIs(c.slack, None)
        self.assertIs(c.lslack, None)
        self.assertIs(c.uslack, None)

        c.r.value = 8
        c.x1.value = 1.1
        c.x2.value = 2.3
        val = round(1.1 * math.exp(2.3 / 1.1) - 8, 9)
        self.assertEqual(round(c(), 9), val)
        self.assertEqual(round(c.slack, 9), -val)
        self.assertEqual(c.lslack, float('inf'))
        self.assertEqual(round(c.uslack, 9), -val)
        self.assertIs(c._body, None)
        # check body
        self.assertEqual(round(c.body(), 9), val)
        self.assertEqual(round(c(), 9), val)
        self.assertEqual(round(c.slack, 9), -val)
        self.assertEqual(c.lslack, float('inf'))
        self.assertEqual(round(c.uslack, 9), -val)
        self.assertIsNot(c._body, None)

    def test_check_convexity_conditions(self):
        c = self._object_factory()
        self.assertEqual(c.check_convexity_conditions(), True)

        c = self._object_factory()
        c.r.domain_type = IntegerSet
        self.assertEqual(c.check_convexity_conditions(), False)
        self.assertEqual(c.check_convexity_conditions(relax=True), True)
        c = self._object_factory()
        c.r.lb = None
        self.assertEqual(c.check_convexity_conditions(), False)
        c = self._object_factory()
        c.r.lb = -1
        self.assertEqual(c.check_convexity_conditions(), False)

        c = self._object_factory()
        c.x1.domain_type = IntegerSet
        self.assertEqual(c.check_convexity_conditions(), False)
        self.assertEqual(c.check_convexity_conditions(relax=True), True)
        c = self._object_factory()
        c.x1.lb = None
        self.assertEqual(c.check_convexity_conditions(), False)
        c = self._object_factory()
        c.x1.lb = -1
        self.assertEqual(c.check_convexity_conditions(), False)

        c = self._object_factory()
        c.x2.domain_type = IntegerSet
        self.assertEqual(c.check_convexity_conditions(), False)
        self.assertEqual(c.check_convexity_conditions(relax=True), True)

    def test_as_domain(self):
        ret = primal_exponential.as_domain(r=3, x1=1, x2=2)
        self.assertIs(type(ret), block)
        q, c, r, x1, x2 = ret.q, ret.c, ret.r, ret.x1, ret.x2
        self.assertEqual(q.check_convexity_conditions(), True)
        self.assertIs(type(q), primal_exponential)
        self.assertIs(type(r), variable)
        self.assertIs(type(x1), variable)
        self.assertIs(type(x2), variable)
        self.assertIs(type(c), constraint_tuple)
        self.assertEqual(len(c), 3)
        self.assertEqual(c[0].rhs, 3)
        r.value = 3
        self.assertEqual(c[0].slack, 0)
        r.value = None
        self.assertEqual(c[1].rhs, 1)
        x1.value = 1
        self.assertEqual(c[1].slack, 0)
        x1.value = None
        self.assertEqual(c[2].rhs, 2)
        x2.value = 2
        self.assertEqual(c[2].slack, 0)
        x2.value = None


class Test_primal_power(_conic_tester_base, unittest.TestCase):
    _object_factory = lambda self: primal_power(
        r1=variable(lb=0),
        r2=variable(lb=0),
        x=[variable(), variable()],
        alpha=parameter(value=0.4),
    )

    def test_bad_alpha_type(self):
        c = primal_power(
            r1=variable(lb=0),
            r2=variable(lb=0),
            x=[variable(), variable()],
            alpha=parameter(),
        )
        c = primal_power(
            r1=variable(lb=0),
            r2=variable(lb=0),
            x=[variable(), variable()],
            alpha=data_expression(),
        )
        with self.assertRaises(TypeError):
            c = primal_power(
                r1=variable(lb=0),
                r2=variable(lb=0),
                x=[variable(), variable()],
                alpha=variable(),
            )
        with self.assertRaises(TypeError):
            c = primal_power(
                r1=variable(lb=0),
                r2=variable(lb=0),
                x=[variable(), variable()],
                alpha=expression(),
            )

    def test_expression(self):
        c = self._object_factory()
        self.assertIs(c._body, None)
        with self.assertRaises(ValueError):
            self.assertIs(c(), None)
        with self.assertRaises(ValueError):
            self.assertIs(c(exception=True), None)
        self.assertIs(c(exception=False), None)
        self.assertIs(c._body, None)
        self.assertIs(c.slack, None)
        self.assertIs(c.lslack, None)
        self.assertIs(c.uslack, None)

        c.r1.value = 5.9
        c.r2.value = 3.4
        c.x[0].value = 1.1
        c.x[1].value = -2.3
        val = round((1.1**2 + (-2.3) ** 2) ** 0.5 - (5.9**0.4) * (3.4**0.6), 9)
        self.assertEqual(round(c(), 9), val)
        self.assertEqual(round(c.slack, 9), -val)
        self.assertEqual(c.lslack, float('inf'))
        self.assertEqual(round(c.uslack, 9), -val)
        self.assertIs(c._body, None)
        # check body
        self.assertEqual(round(c.body(), 9), val)
        self.assertEqual(round(c(), 9), val)
        self.assertEqual(round(c.slack, 9), -val)
        self.assertEqual(c.lslack, float('inf'))
        self.assertEqual(round(c.uslack, 9), -val)
        self.assertIsNot(c._body, None)

    def test_check_convexity_conditions(self):
        c = self._object_factory()
        self.assertEqual(c.check_convexity_conditions(), True)

        c = self._object_factory()
        c.r1.domain_type = IntegerSet
        self.assertEqual(c.check_convexity_conditions(), False)
        self.assertEqual(c.check_convexity_conditions(relax=True), True)
        c = self._object_factory()
        c.r1.lb = None
        self.assertEqual(c.check_convexity_conditions(), False)
        c = self._object_factory()
        c.r1.lb = -1
        self.assertEqual(c.check_convexity_conditions(), False)

        c = self._object_factory()
        c.r2.domain_type = IntegerSet
        self.assertEqual(c.check_convexity_conditions(), False)
        self.assertEqual(c.check_convexity_conditions(relax=True), True)
        c = self._object_factory()
        c.r2.lb = None
        self.assertEqual(c.check_convexity_conditions(), False)
        c = self._object_factory()
        c.r2.lb = -1
        self.assertEqual(c.check_convexity_conditions(), False)

        c = self._object_factory()
        c.x[0].domain_type = IntegerSet
        self.assertEqual(c.check_convexity_conditions(), False)
        self.assertEqual(c.check_convexity_conditions(relax=True), True)

        c = self._object_factory()
        c.alpha.value = 0
        self.assertEqual(c.check_convexity_conditions(), False)
        c = self._object_factory()
        c.alpha.value = 1
        self.assertEqual(c.check_convexity_conditions(), False)

    def test_as_domain(self):
        ret = primal_power.as_domain(r1=3, r2=4, x=[1, 2], alpha=0.5)
        self.assertIs(type(ret), block)
        q, c, r1, r2, x = ret.q, ret.c, ret.r1, ret.r2, ret.x
        self.assertEqual(q.check_convexity_conditions(), True)
        self.assertIs(type(q), primal_power)
        self.assertIs(type(r1), variable)
        self.assertIs(type(r2), variable)
        self.assertIs(type(x), variable_tuple)
        self.assertEqual(len(x), 2)
        self.assertIs(type(c), constraint_tuple)
        self.assertEqual(len(c), 4)
        self.assertEqual(c[0].rhs, 3)
        r1.value = 3
        self.assertEqual(c[0].slack, 0)
        r1.value = None
        self.assertEqual(c[1].rhs, 4)
        r2.value = 4
        self.assertEqual(c[1].slack, 0)
        r2.value = None
        self.assertEqual(c[2].rhs, 1)
        x[0].value = 1
        self.assertEqual(c[2].slack, 0)
        x[0].value = None
        self.assertEqual(c[3].rhs, 2)
        x[1].value = 2
        self.assertEqual(c[3].slack, 0)
        x[1].value = None


class Test_dual_exponential(_conic_tester_base, unittest.TestCase):
    _object_factory = lambda self: dual_exponential(
        r=variable(lb=0), x1=variable(), x2=variable(ub=0)
    )

    def test_expression(self):
        c = self._object_factory()
        self.assertIs(c._body, None)
        with self.assertRaises(ValueError):
            self.assertIs(c(), None)
        with self.assertRaises(ValueError):
            self.assertIs(c(exception=True), None)
        self.assertIs(c(exception=False), None)
        self.assertIs(c._body, None)
        self.assertIs(c.slack, None)
        self.assertIs(c.lslack, None)
        self.assertIs(c.uslack, None)

        c.r.value = 2.7
        c.x1.value = 1.2
        c.x2.value = -5.3
        val = round(-(-5.3 / math.e) * math.exp(1.2 / -5.3) - 2.7, 9)
        self.assertEqual(round(c(), 9), val)
        self.assertEqual(round(c.slack, 9), -val)
        self.assertEqual(c.lslack, float('inf'))
        self.assertEqual(round(c.uslack, 9), -val)
        self.assertIs(c._body, None)
        # check body
        self.assertEqual(round(c.body(), 9), val)
        self.assertEqual(round(c(), 9), val)
        self.assertEqual(round(c.slack, 9), -val)
        self.assertEqual(c.lslack, float('inf'))
        self.assertEqual(round(c.uslack, 9), -val)
        self.assertIsNot(c._body, None)

    def test_check_convexity_conditions(self):
        c = self._object_factory()
        self.assertEqual(c.check_convexity_conditions(), True)

        c = self._object_factory()
        c.r.domain_type = IntegerSet
        self.assertEqual(c.check_convexity_conditions(), False)
        self.assertEqual(c.check_convexity_conditions(relax=True), True)
        c = self._object_factory()
        c.r.lb = None
        self.assertEqual(c.check_convexity_conditions(), False)
        c = self._object_factory()
        c.r.lb = -1
        self.assertEqual(c.check_convexity_conditions(), False)

        c = self._object_factory()
        c.x1.domain_type = IntegerSet
        self.assertEqual(c.check_convexity_conditions(), False)
        self.assertEqual(c.check_convexity_conditions(relax=True), True)

        c = self._object_factory()
        c.x2.domain_type = IntegerSet
        self.assertEqual(c.check_convexity_conditions(), False)
        self.assertEqual(c.check_convexity_conditions(relax=True), True)
        c = self._object_factory()
        c.x2.ub = None
        self.assertEqual(c.check_convexity_conditions(), False)
        c = self._object_factory()
        c.x2.ub = 1
        self.assertEqual(c.check_convexity_conditions(), False)

    def test_as_domain(self):
        ret = dual_exponential.as_domain(r=3, x1=1, x2=2)
        self.assertIs(type(ret), block)
        q, c, r, x1, x2 = ret.q, ret.c, ret.r, ret.x1, ret.x2
        self.assertEqual(q.check_convexity_conditions(), True)
        self.assertIs(type(q), dual_exponential)
        self.assertIs(type(x1), variable)
        self.assertIs(type(x2), variable)
        self.assertIs(type(r), variable)
        self.assertIs(type(c), constraint_tuple)
        self.assertEqual(len(c), 3)
        self.assertEqual(c[0].rhs, 3)
        r.value = 3
        self.assertEqual(c[0].slack, 0)
        r.value = None
        self.assertEqual(c[1].rhs, 1)
        x1.value = 1
        self.assertEqual(c[1].slack, 0)
        x1.value = None
        self.assertEqual(c[2].rhs, 2)
        x2.value = 2
        self.assertEqual(c[2].slack, 0)
        x2.value = None


class Test_dual_power(_conic_tester_base, unittest.TestCase):
    _object_factory = lambda self: dual_power(
        r1=variable(lb=0),
        r2=variable(lb=0),
        x=[variable(), variable()],
        alpha=parameter(value=0.4),
    )

    def test_bad_alpha_type(self):
        c = dual_power(
            r1=variable(lb=0),
            r2=variable(lb=0),
            x=[variable(), variable()],
            alpha=parameter(),
        )
        c = dual_power(
            r1=variable(lb=0),
            r2=variable(lb=0),
            x=[variable(), variable()],
            alpha=data_expression(),
        )
        with self.assertRaises(TypeError):
            c = dual_power(
                r1=variable(lb=0),
                r2=variable(lb=0),
                x=[variable(), variable()],
                alpha=variable(),
            )
        with self.assertRaises(TypeError):
            c = dual_power(
                r1=variable(lb=0),
                r2=variable(lb=0),
                x=[variable(), variable()],
                alpha=expression(),
            )

    def test_expression(self):
        c = self._object_factory()
        self.assertIs(c._body, None)
        with self.assertRaises(ValueError):
            self.assertIs(c(), None)
        with self.assertRaises(ValueError):
            self.assertIs(c(exception=True), None)
        self.assertIs(c(exception=False), None)
        self.assertIs(c._body, None)
        self.assertIs(c.slack, None)
        self.assertIs(c.lslack, None)
        self.assertIs(c.uslack, None)

        c.r1.value = 2.7
        c.r2.value = 3.7
        c.x[0].value = 1.2
        c.x[1].value = -5.3
        val = round(
            (1.2**2 + (-5.3) ** 2) ** 0.5 - ((2.7 / 0.4) ** 0.4) * ((3.7 / 0.6) ** 0.6),
            9,
        )
        self.assertEqual(round(c(), 9), val)
        self.assertEqual(round(c.slack, 9), -val)
        self.assertEqual(c.lslack, float('inf'))
        self.assertEqual(round(c.uslack, 9), -val)
        self.assertIs(c._body, None)
        # check body
        self.assertEqual(round(c.body(), 9), val)
        self.assertEqual(round(c(), 9), val)
        self.assertEqual(round(c.slack, 9), -val)
        self.assertEqual(c.lslack, float('inf'))
        self.assertEqual(round(c.uslack, 9), -val)
        self.assertIsNot(c._body, None)

    def test_check_convexity_conditions(self):
        c = self._object_factory()
        self.assertEqual(c.check_convexity_conditions(), True)

        c = self._object_factory()
        c.r1.domain_type = IntegerSet
        self.assertEqual(c.check_convexity_conditions(), False)
        self.assertEqual(c.check_convexity_conditions(relax=True), True)
        c = self._object_factory()
        c.r1.lb = None
        self.assertEqual(c.check_convexity_conditions(), False)
        c = self._object_factory()
        c.r1.lb = -1
        self.assertEqual(c.check_convexity_conditions(), False)

        c = self._object_factory()
        c.r2.domain_type = IntegerSet
        self.assertEqual(c.check_convexity_conditions(), False)
        self.assertEqual(c.check_convexity_conditions(relax=True), True)
        c = self._object_factory()
        c.r2.lb = None
        self.assertEqual(c.check_convexity_conditions(), False)
        c = self._object_factory()
        c.r2.lb = -1
        self.assertEqual(c.check_convexity_conditions(), False)

        c = self._object_factory()
        c.x[0].domain_type = IntegerSet
        self.assertEqual(c.check_convexity_conditions(), False)
        self.assertEqual(c.check_convexity_conditions(relax=True), True)

        c = self._object_factory()
        c.alpha.value = 0
        self.assertEqual(c.check_convexity_conditions(), False)
        c = self._object_factory()
        c.alpha.value = 1
        self.assertEqual(c.check_convexity_conditions(), False)

    def test_as_domain(self):
        ret = dual_power.as_domain(r1=3, r2=4, x=[1, 2], alpha=0.5)
        self.assertIs(type(ret), block)
        q, c, r1, r2, x = ret.q, ret.c, ret.r1, ret.r2, ret.x
        self.assertEqual(q.check_convexity_conditions(), True)
        self.assertIs(type(q), dual_power)
        self.assertIs(type(r1), variable)
        self.assertIs(type(r2), variable)
        self.assertIs(type(x), variable_tuple)
        self.assertEqual(len(x), 2)
        self.assertIs(type(c), constraint_tuple)
        self.assertEqual(len(c), 4)
        self.assertEqual(c[0].rhs, 3)
        r1.value = 3
        self.assertEqual(c[0].slack, 0)
        r1.value = None
        self.assertEqual(c[1].rhs, 4)
        r2.value = 4
        self.assertEqual(c[1].slack, 0)
        r2.value = None
        self.assertEqual(c[2].rhs, 1)
        x[0].value = 1
        self.assertEqual(c[2].slack, 0)
        x[0].value = None
        self.assertEqual(c[3].rhs, 2)
        x[1].value = 2
        self.assertEqual(c[3].slack, 0)
        x[1].value = None


# These mosek 10 constraints can't be evaluated, pprinted, checked for convexity,
# pickled, etc., so I won't use the _conic_tester_base for them
class Test_primal_geomean(unittest.TestCase):
    def test_as_domain(self):
        b = primal_geomean.as_domain(r=[2, 3], x=6)
        self.assertIs(type(b), block)
        self.assertIs(type(b.q), primal_geomean)
        self.assertIs(type(b.r), variable_tuple)
        self.assertIs(type(b.x), variable)
        self.assertIs(type(b.c), constraint_tuple)
        self.assertExpressionsEqual(b.c[0].body, b.r[0])
        self.assertExpressionsEqual(b.c[0].rhs, 2)
        self.assertExpressionsEqual(b.c[1].body, b.r[1])
        self.assertExpressionsEqual(b.c[1].rhs, 3)
        self.assertExpressionsEqual(b.c[2].body, b.x)
        self.assertExpressionsEqual(b.c[2].rhs, 6)


class Test_dual_geomean(unittest.TestCase):
    def test_as_domain(self):
        b = dual_geomean.as_domain(r=[2, 3], x=6)
        self.assertIs(type(b), block)
        self.assertIs(type(b.q), dual_geomean)
        self.assertIs(type(b.r), variable_tuple)
        self.assertIs(type(b.x), variable)
        self.assertIs(type(b.c), constraint_tuple)
        self.assertExpressionsEqual(b.c[0].body, b.r[0])
        self.assertExpressionsEqual(b.c[0].rhs, 2)
        self.assertExpressionsEqual(b.c[1].body, b.r[1])
        self.assertExpressionsEqual(b.c[1].rhs, 3)
        self.assertExpressionsEqual(b.c[2].body, b.x)
        self.assertExpressionsEqual(b.c[2].rhs, 6)


class TestMisc(unittest.TestCase):
    def test_build_linking_constraints(self):
        c = _build_linking_constraints([], [])
        self.assertIs(type(c), constraint_tuple)
        self.assertEqual(len(c), 0)
        c = _build_linking_constraints([None], [variable()])
        self.assertIs(type(c), constraint_tuple)
        self.assertEqual(len(c), 0)
        v = [1, data_expression(), variable(), expression(expr=1.0)]
        vaux = [variable(), variable(), variable(), variable()]
        c = _build_linking_constraints(v, vaux)
        self.assertIs(type(c), constraint_tuple)
        self.assertEqual(len(c), 4)
        self.assertIs(type(c[0]), linear_constraint)
        self.assertEqual(c[0].rhs, 1)
        self.assertEqual(len(list(c[0].terms)), 1)
        self.assertIs(list(c[0].terms)[0][0], vaux[0])
        self.assertEqual(list(c[0].terms)[0][1], 1)
        self.assertIs(type(c[1]), linear_constraint)
        self.assertIs(c[1].rhs, v[1])
        self.assertEqual(len(list(c[1].terms)), 1)
        self.assertIs(list(c[1].terms)[0][0], vaux[1])
        self.assertEqual(list(c[1].terms)[0][1], 1)
        self.assertIs(type(c[2]), linear_constraint)
        self.assertEqual(c[2].rhs, 0)
        self.assertEqual(len(list(c[2].terms)), 2)
        self.assertIs(list(c[2].terms)[0][0], vaux[2])
        self.assertEqual(list(c[2].terms)[0][1], 1)
        self.assertIs(list(c[2].terms)[1][0], v[2])
        self.assertEqual(list(c[2].terms)[1][1], -1)
        self.assertIs(type(c[3]), constraint)
        self.assertEqual(c[3].rhs, 0)
        from pyomo.repn import generate_standard_repn

        repn = generate_standard_repn(c[3].body)
        self.assertEqual(len(repn.linear_vars), 1)
        self.assertIs(repn.linear_vars[0], vaux[3])
        self.assertEqual(repn.linear_coefs[0], 1)
        self.assertEqual(repn.constant, -1)


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

#
# AMPL Problem Writer Plugin
#

import itertools
import logging
import operator
import os
import time
from math import isclose

from pyomo.common.fileutils import find_library
from pyomo.common.gc_manager import PauseGC
from pyomo.opt import ProblemFormat, AbstractProblemWriter, WriterFactory
import pyomo.core.expr as EXPR
from pyomo.core.expr.numvalue import (
    NumericConstant,
    native_numeric_types,
    value,
    is_fixed,
)
from pyomo.core.base import (
    SymbolMap,
    NameLabeler,
    NamedExpressionData,
    SortComponents,
    var,
    param,
    Var,
    ExternalFunction,
    ComponentMap,
    Objective,
    Constraint,
    SOSConstraint,
    Suffix,
)
import pyomo.core.base.suffix
from pyomo.repn.standard_repn import generate_standard_repn

import pyomo.core.kernel.suffix
from pyomo.core.kernel.block import IBlock
from pyomo.core.kernel.expression import IIdentityExpression
from pyomo.core.kernel.variable import IVariable

logger = logging.getLogger('pyomo.core')


def set_pyomo_amplfunc_env(external_libs):
    # The ASL AMPLFUNC environment variable is nominally a
    # whitespace-separated string of library names.  Beginning
    # sometime between 2010 and 2012, the ASL added support for
    # simple quoted strings: the first non-whitespace character
    # can be either " or '.  When that is detected, the ASL
    # parser will continue to the next occurrence of that
    # character (i.e., no escaping is allowed).  We will use
    # that same logic here to quote any strings with spaces
    # ... bearing in mind that this will only work with solvers
    # compiled against versions of the ASL more recent than
    # ~2012.
    #
    # We are (arbitrarily) choosing to use newline as the field
    # separator.
    env_str = ''
    for _lib in external_libs:
        _lib = _lib.strip()
        # Convert the library to an absolute path
        _abs_lib = find_library(_lib)
        if _abs_lib is not None:
            _lib = _abs_lib
        if (
            ' ' not in _lib
            or (_lib[0] == '"' and _lib[-1] == '"' and '"' not in _lib[1:-1])
            or (_lib[0] == "'" and _lib[-1] == "'" and "'" not in _lib[1:-1])
        ):
            pass
        elif '"' not in _lib:
            _lib = '"' + _lib + '"'
        elif "'" not in _lib:
            _lib = "'" + _lib + "'"
        else:
            raise RuntimeError(
                "Cannot pass the AMPL external function library\n\t%s\n"
                "to the ASL because the string contains spaces, "
                "single quote and\ndouble quote characters." % (_lib,)
            )
        if env_str:
            env_str += "\n"
        env_str += _lib
    os.environ["PYOMO_AMPLFUNC"] = env_str


_intrinsic_function_operators = {
    'log': 'o43',
    'log10': 'o42',
    'sin': 'o41',
    'cos': 'o46',
    'tan': 'o38',
    'sinh': 'o40',
    'cosh': 'o45',
    'tanh': 'o37',
    'asin': 'o51',
    'acos': 'o53',
    'atan': 'o49',
    'exp': 'o44',
    'sqrt': 'o39',
    'asinh': 'o50',
    'acosh': 'o52',
    'atanh': 'o47',
    'pow': 'o5',
    'abs': 'o15',
    'ceil': 'o14',
    'floor': 'o13',
}


# build string templates
def _build_op_template():
    _op_template = {}
    _op_comment = {}

    prod_template = "o2{C}\n"
    prod_comment = "\t#*"
    div_template = "o3{C}\n"
    div_comment = "\t#/"
    _op_template[EXPR.ProductExpression] = prod_template
    _op_comment[EXPR.ProductExpression] = prod_comment
    _op_template[EXPR.DivisionExpression] = div_template
    _op_comment[EXPR.DivisionExpression] = div_comment

    _op_template[EXPR.ExternalFunctionExpression] = (
        "f%d %d{C}\n",  # function
        "h%d:%s{C}\n",
    )  # string arg
    _op_comment[EXPR.ExternalFunctionExpression] = (
        "\t#%s",  # function
        "",
    )  # string arg

    for opname in _intrinsic_function_operators:
        _op_template[opname] = _intrinsic_function_operators[opname] + "{C}\n"
        _op_comment[opname] = "\t#" + opname

    _op_template[EXPR.Expr_ifExpression] = "o35{C}\n"
    _op_comment[EXPR.Expr_ifExpression] = "\t#if"

    _op_template[EXPR.InequalityExpression] = (
        "o21{C}\n",  # and
        "o22{C}\n",  # <
        "o23{C}\n",
    )  # <=
    _op_comment[EXPR.InequalityExpression] = (
        "\t#and",  # and
        "\t#lt",  # <
        "\t#le",
    )  # <=

    _op_template[EXPR.EqualityExpression] = "o24{C}\n"
    _op_comment[EXPR.EqualityExpression] = "\t#eq"

    _op_template[var.VarData] = "v%d{C}\n"
    _op_comment[var.VarData] = "\t#%s"

    _op_template[param.ParamData] = "n%r{C}\n"
    _op_comment[param.ParamData] = ""

    _op_template[NumericConstant] = "n%r{C}\n"
    _op_comment[NumericConstant] = ""

    _op_template[EXPR.SumExpressionBase] = (
        "o54{C}\n%d\n",  # nary +
        "o0{C}\n",  # +
        "o2\n" + _op_template[NumericConstant],
    )  # * coef
    _op_comment[EXPR.SumExpressionBase] = (
        "\t#sumlist",  # nary +
        "\t#+",  # +
        _op_comment[NumericConstant],
    )  # * coef
    _op_template[EXPR.NegationExpression] = "o16{C}\n"
    _op_comment[EXPR.NegationExpression] = "\t#-"

    return _op_template, _op_comment


def _get_bound(exp):
    if exp is None:
        return None
    if is_fixed(exp):
        return value(exp)
    raise ValueError("non-fixed bound or weight: " + str(exp))


class StopWatch:
    def __init__(self):
        self.start = time.time()

    def report(self, msg):
        print(msg + " (seconds): " + str(time.time() - self.start))

    def reset(self):
        self.start = time.time()


class _Counter:
    def __init__(self, start):
        self._id = start

    def __call__(self, obj):
        tmp = self._id
        self._id += 1
        return tmp


class ModelSOS:
    class AmplSuffix:
        def __init__(self, name):
            self.name = name
            self.ids = []
            self.vals = []

        def add(self, idx, val):
            if idx in self.ids:
                raise RuntimeError(
                    "The NL file format does not support multiple nonzero "
                    "values for a single component and suffix. \n"
                    "Suffix Name:  %s\n"
                    "Component ID: %s\n" % (self.name, idx)
                )
            else:
                self.ids.append(idx)
                self.vals.append(val)

        def genfilelines(self):
            base_line = "{0} {1}\n"
            return [
                base_line.format(idx, val)
                for idx, val in zip(self.ids, self.vals)
                if val != 0
            ]

        def is_empty(self):
            return not bool(len(self.ids))

    def __init__(self, ampl_var_id, varID_map):
        self.ampl_var_id = ampl_var_id
        self.sosno = self.AmplSuffix('sosno')
        self.ref = self.AmplSuffix('ref')
        self.block_cntr = 0
        self.varID_map = varID_map

    def count_constraint(self, soscondata):
        ampl_var_id = self.ampl_var_id
        varID_map = self.varID_map

        if hasattr(soscondata, 'get_items'):
            sos_items = list(soscondata.get_items())
        else:
            sos_items = list(soscondata.items())

        if len(sos_items) == 0:
            return

        level = soscondata.level

        # Identifies the the current set from others in the NL file
        self.block_cntr += 1

        # If SOS1, the identifier must be positive
        # if SOS2, the identifier must be negative
        sign_tag = None
        if level == 1:
            sign_tag = 1
        elif level == 2:
            sign_tag = -1
        else:
            raise ValueError(
                "SOSConstraint '%s' has sos type='%s', "
                "which is not supported by the NL file interface"
                % (soscondata.name, level)
            )

        for vardata, weight in sos_items:
            weight = _get_bound(weight)
            if weight < 0:
                raise ValueError(
                    "Cannot use negative weight %f "
                    "for variable %s is special ordered "
                    "set %s " % (weight, vardata.name, soscondata.name)
                )
            if vardata.fixed:
                raise ValueError(
                    "SOSConstraint '%s' includes a fixed Variable '%s'. "
                    "This is currently not supported. Deactivate this constraint "
                    "in order to proceed" % (soscondata.name, vardata.name)
                )

            ID = ampl_var_id[varID_map[id(vardata)]]
            self.sosno.add(ID, self.block_cntr * sign_tag)
            self.ref.add(ID, weight)


class RepnWrapper:
    __slots__ = ('repn', 'linear_vars', 'nonlinear_vars')

    def __init__(self, repn, linear, nonlinear):
        self.repn = repn
        self.linear_vars = linear
        self.nonlinear_vars = nonlinear


@WriterFactory.register('nl_v1', 'Generate the corresponding AMPL NL file (version 1).')
class ProblemWriter_nl(AbstractProblemWriter):
    def __init__(self):
        AbstractProblemWriter.__init__(self, ProblemFormat.nl)
        self._ampl_var_id = {}
        self._ampl_con_id = {}
        self._ampl_obj_id = {}
        self._OUTPUT = None
        self._varID_map = None

    def __call__(self, model, filename, solver_capability, io_options):
        # Rebuild the OP template (as the expression tree system may
        # have been switched)
        _op_template, _op_comment = _build_op_template()

        # Make sure not to modify the user's dictionary, they may be
        # reusing it outside of this call
        io_options = dict(io_options)

        # NOTE: io_options is a simple dictionary of keyword-value pairs
        #       specific to this writer. that said, we are not good
        #       about enforcing consistency between the io_options and
        #       kwds - arguably, we should go with the former or the
        #       latter, but not support both. then again, the kwds are
        #       not used anywhere within this function.

        # Print timing after writing each section of the NL file
        show_section_timing = io_options.pop("show_section_timing", False)

        # Skip writing constraints whose body section is fixed (i.e., no variables)
        skip_trivial_constraints = io_options.pop("skip_trivial_constraints", False)

        # How much effort do we want to put into ensuring the
        # NL file is written deterministically for a Pyomo model:
        #    0 : None
        #    1 : sort keys of indexed components (default)
        #    2 : sort keys AND sort names (over declaration order)
        file_determinism = io_options.pop("file_determinism", 1)

        # Write the corresponding .row and .col files for the NL files
        # identifying variable and constraint indices in the NLP
        # matrix.
        symbolic_solver_labels = io_options.pop("symbolic_solver_labels", False)

        # If False, we raise an error if fixed variables are
        # encountered in the list of active variables (i.e., an
        # expression containing some fixed variable was not
        # preprocessed after the variable was fixed). If True, we
        # allow this case and modify the variable bounds section to
        # fix the variable.
        output_fixed_variable_bounds = io_options.pop(
            "output_fixed_variable_bounds", False
        )

        # If False, unused variables will not be included in
        # the NL file. Otherwise, include all variables in
        # the bounds sections.
        include_all_variable_bounds = io_options.pop(
            "include_all_variable_bounds", False
        )

        # List of variables that don't appear in constraints to force into the
        # nl-file
        export_nonlinear_variables = io_options.pop("export_nonlinear_variables", False)

        # column_order is a new option supported by the nl writer v2
        _column_order = io_options.pop("column_order", True)
        assert _column_order in {True}

        if len(io_options):
            raise ValueError(
                "ProblemWriter_nl passed unrecognized io_options:\n\t"
                + "\n\t".join("%s = %s" % (k, v) for k, v in io_options.items())
            )

        if filename is None:
            filename = model.name + ".nl"

        # Generate the operator strings templates. The value of
        # symbolic_solver_labels determines whether or not to
        # include "nl comments" (the equivalent AMPL functionality
        # is "option nl_comments 1").
        self._op_string = {}
        for optype in _op_template:
            template_str = _op_template[optype]
            comment_str = _op_comment[optype]
            if type(template_str) is tuple:
                op_strings = []
                for i in range(len(template_str)):
                    if symbolic_solver_labels:
                        op_strings.append(template_str[i].format(C=comment_str[i]))
                    else:
                        op_strings.append(template_str[i].format(C=""))
                self._op_string[optype] = tuple(op_strings)
            else:
                if symbolic_solver_labels:
                    self._op_string[optype] = template_str.format(C=comment_str)
                else:
                    self._op_string[optype] = template_str.format(C="")

        # making these attributes so they do not need to be
        # passed into _print_nonlinear_terms_NL
        self._symbolic_solver_labels = symbolic_solver_labels
        self._output_fixed_variable_bounds = output_fixed_variable_bounds
        # Speeds up calling name on every component when
        # writing .row and .col files (when symbolic_solver_labels is True)
        self._name_labeler = NameLabeler()

        # Pause the GC for the duration of this method
        with PauseGC() as pgc:
            with open(filename, "w") as f:
                self._OUTPUT = f
                symbol_map = self._print_model_NL(
                    model,
                    solver_capability,
                    show_section_timing=show_section_timing,
                    skip_trivial_constraints=skip_trivial_constraints,
                    file_determinism=file_determinism,
                    include_all_variable_bounds=include_all_variable_bounds,
                    export_nonlinear_variables=export_nonlinear_variables,
                )

        self._symbolic_solver_labels = False
        self._output_fixed_variable_bounds = False
        self._name_labeler = None

        self._OUTPUT = None
        self._varID_map = None
        self._op_string = None
        return filename, symbol_map

    def _print_quad_term(self, v1, v2):
        OUTPUT = self._OUTPUT
        if v1 is not v2:
            prod_str = self._op_string[EXPR.ProductExpression]
            OUTPUT.write(prod_str)
            self._print_nonlinear_terms_NL(v1)
            self._print_nonlinear_terms_NL(v2)
        else:
            intr_expr_str = self._op_string['pow']
            OUTPUT.write(intr_expr_str)
            self._print_nonlinear_terms_NL(v1)
            OUTPUT.write(self._op_string[NumericConstant] % (2))

    def _print_standard_quadratic_NL(self, quadratic_vars, quadratic_coefs):
        OUTPUT = self._OUTPUT
        nary_sum_str, binary_sum_str, coef_term_str = self._op_string[
            EXPR.SumExpressionBase
        ]
        assert len(quadratic_vars) == len(quadratic_coefs)
        if len(quadratic_vars) == 1:
            pass
        else:
            if len(quadratic_vars) == 2:
                OUTPUT.write(binary_sum_str)
            else:
                assert len(quadratic_vars) > 2
                OUTPUT.write(nary_sum_str % (len(quadratic_vars)))
            # now we need to do a sort to ensure deterministic output
            # as the compiled quadratic representation does not preserve
            # any ordering
            old_quadratic_vars = quadratic_vars
            old_quadratic_coefs = quadratic_coefs
            self_varID_map = self._varID_map
            quadratic_vars = []
            quadratic_coefs = []
            for i, (v1, v2) in sorted(
                enumerate(old_quadratic_vars),
                key=lambda x: (
                    self_varID_map[id(x[1][0])],
                    self_varID_map[id(x[1][1])],
                ),
            ):
                quadratic_coefs.append(old_quadratic_coefs[i])
                if self_varID_map[id(v1)] <= self_varID_map[id(v2)]:
                    quadratic_vars.append((v1, v2))
                else:
                    quadratic_vars.append((v2, v1))
        for i in range(len(quadratic_vars)):
            coef = quadratic_coefs[i]
            v1, v2 = quadratic_vars[i]
            if coef != 1:
                OUTPUT.write(coef_term_str % (coef))
            self._print_quad_term(v1, v2)

    def _print_nonlinear_terms_NL(self, exp):
        OUTPUT = self._OUTPUT
        exp_type = type(exp)
        # JDS: check list first so that after this, we know that exp
        # must be some form of NumericValue
        if exp_type is list:
            # this is an implied summation of expressions (we did not
            # create a new sum expression for efficiency) this should
            # be a list of tuples where [0] is the coeff and [1] is
            # the expr to write
            nary_sum_str, binary_sum_str, coef_term_str = self._op_string[
                EXPR.SumExpressionBase
            ]
            n = len(exp)
            if n > 2:
                OUTPUT.write(nary_sum_str % (n))
                for i in range(0, n):
                    assert exp[i].__class__ is tuple
                    coef = exp[i][0]
                    child_exp = exp[i][1]
                    if coef != 1:
                        OUTPUT.write(coef_term_str % (coef))
                    self._print_nonlinear_terms_NL(child_exp)
            else:  # n == 1 or 2
                for i in range(0, n):
                    assert exp[i].__class__ is tuple
                    coef = exp[i][0]
                    child_exp = exp[i][1]
                    if i != n - 1:
                        # need the + op if it is not the last entry in the list
                        OUTPUT.write(binary_sum_str)
                    if coef != 1:
                        OUTPUT.write(coef_term_str % (coef))
                    self._print_nonlinear_terms_NL(child_exp)
        elif exp_type in native_numeric_types:
            OUTPUT.write(self._op_string[NumericConstant] % (exp))

        elif exp.is_expression_type():
            #
            # Identify NPV expressions
            #
            if not exp.is_potentially_variable():
                OUTPUT.write(self._op_string[NumericConstant] % (value(exp)))
            #
            # We are assuming that _Constant_* expression objects
            # have been preprocessed to form constant values.
            #
            elif (
                exp.__class__ is EXPR.SumExpression
                or exp.__class__ is EXPR.LinearExpression
            ):
                nary_sum_str, binary_sum_str, coef_term_str = self._op_string[
                    EXPR.SumExpressionBase
                ]
                n = exp.nargs()
                const = 0
                vargs = []
                for v in exp.args:
                    if v.__class__ in native_numeric_types:
                        const += v
                    else:
                        vargs.append(v)
                if not isclose(const, 0.0):
                    vargs.append(const)
                n = len(vargs)
                if n == 2:
                    OUTPUT.write(binary_sum_str)
                    self._print_nonlinear_terms_NL(vargs[0])
                    self._print_nonlinear_terms_NL(vargs[1])
                elif n == 1:
                    self._print_nonlinear_terms_NL(vargs[0])
                else:
                    OUTPUT.write(nary_sum_str % (n))
                    for child_exp in vargs:
                        self._print_nonlinear_terms_NL(child_exp)

            elif exp_type is EXPR.SumExpressionBase:
                nary_sum_str, binary_sum_str, coef_term_str = self._op_string[
                    EXPR.SumExpressionBase
                ]
                OUTPUT.write(binary_sum_str)
                self._print_nonlinear_terms_NL(exp.arg(0))
                self._print_nonlinear_terms_NL(exp.arg(1))

            elif exp_type is EXPR.MonomialTermExpression:
                prod_str = self._op_string[EXPR.ProductExpression]
                OUTPUT.write(prod_str)
                self._print_nonlinear_terms_NL(value(exp.arg(0)))
                self._print_nonlinear_terms_NL(exp.arg(1))

            elif exp_type is EXPR.ProductExpression:
                prod_str = self._op_string[EXPR.ProductExpression]
                OUTPUT.write(prod_str)
                self._print_nonlinear_terms_NL(exp.arg(0))
                self._print_nonlinear_terms_NL(exp.arg(1))

            elif exp_type is EXPR.DivisionExpression:
                assert exp.nargs() == 2
                div_str = self._op_string[EXPR.DivisionExpression]
                OUTPUT.write(div_str)
                self._print_nonlinear_terms_NL(exp.arg(0))
                self._print_nonlinear_terms_NL(exp.arg(1))

            elif exp_type is EXPR.NegationExpression:
                assert exp.nargs() == 1
                OUTPUT.write(self._op_string[EXPR.NegationExpression])
                self._print_nonlinear_terms_NL(exp.arg(0))

            elif exp_type is EXPR.ExternalFunctionExpression:
                # We have found models where external functions with
                # strictly fixed/constant arguments causes AMPL to
                # SEGFAULT.  To be safe, we will collapse fixed
                # arguments to scalars and if the entire expression is
                # constant, we will eliminate the external function
                # call entirely.
                if exp.is_fixed():
                    self._print_nonlinear_terms_NL(exp())
                    return
                fun_str, string_arg_str = self._op_string[
                    EXPR.ExternalFunctionExpression
                ]
                if not self._symbolic_solver_labels:
                    OUTPUT.write(
                        fun_str
                        % (self.external_byFcn[exp._fcn._function][1], exp.nargs())
                    )
                else:
                    # Note: exp.name fails
                    OUTPUT.write(
                        fun_str
                        % (
                            self.external_byFcn[exp._fcn._function][1],
                            exp.nargs(),
                            exp.name,
                        )
                    )
                for arg in exp.args:
                    if isinstance(arg, str):
                        # Note: ASL does not handle '\r\n' as the EOL
                        # marker for string arguments (even though it
                        # allows them elsewhere in the NL file).
                        # Since we know we opened this file object (and
                        # that it points to an actual file), we will
                        # grab the underlying fileno and write to it
                        # WITHOUT universal newlines.
                        #
                        # We could just switch the NL writer to always
                        # write out UNIX-style line endings, but that
                        # would force us to change a large number of
                        # baselines / file comparisons
                        OUTPUT.flush()
                        with os.fdopen(
                            OUTPUT.fileno(), mode='w+', closefd=False, newline='\n'
                        ) as TMP:
                            TMP.write(string_arg_str % (len(arg), arg))
                    elif type(arg) in native_numeric_types:
                        self._print_nonlinear_terms_NL(arg)
                    elif arg.is_fixed():
                        self._print_nonlinear_terms_NL(arg())
                    else:
                        self._print_nonlinear_terms_NL(arg)

            elif exp_type is EXPR.PowExpression:
                intr_expr_str = self._op_string['pow']
                OUTPUT.write(intr_expr_str)
                self._print_nonlinear_terms_NL(exp.arg(0))
                self._print_nonlinear_terms_NL(exp.arg(1))

            elif isinstance(exp, EXPR.UnaryFunctionExpression):
                assert exp.nargs() == 1
                intr_expr_str = self._op_string.get(exp.name)
                if intr_expr_str is not None:
                    OUTPUT.write(intr_expr_str)
                else:
                    logger.error("Unsupported unary function ({0})".format(exp.name))
                    raise TypeError(
                        "ASL writer does not support '%s' expressions" % (exp.name)
                    )
                self._print_nonlinear_terms_NL(exp.arg(0))

            elif exp_type is EXPR.Expr_ifExpression:
                OUTPUT.write(self._op_string[EXPR.Expr_ifExpression])
                for arg in exp.args:
                    self._print_nonlinear_terms_NL(arg)

            elif exp_type is EXPR.InequalityExpression:
                and_str, lt_str, le_str = self._op_string[EXPR.InequalityExpression]
                left = exp.arg(0)
                right = exp.arg(1)
                if exp._strict:
                    OUTPUT.write(lt_str)
                else:
                    OUTPUT.write(le_str)
                self._print_nonlinear_terms_NL(left)
                self._print_nonlinear_terms_NL(right)

            elif exp_type is EXPR.RangedExpression:
                and_str, lt_str, le_str = self._op_string[EXPR.InequalityExpression]
                left = exp.arg(0)
                middle = exp.arg(1)
                right = exp.arg(2)
                OUTPUT.write(and_str)
                if exp._strict[0]:
                    OUTPUT.write(lt_str)
                else:
                    OUTPUT.write(le_str)
                self._print_nonlinear_terms_NL(left)
                self._print_nonlinear_terms_NL(middle)
                if exp._strict[1]:
                    OUTPUT.write(lt_str)
                else:
                    OUTPUT.write(le_str)
                self._print_nonlinear_terms_NL(middle)
                self._print_nonlinear_terms_NL(right)

            elif exp_type is EXPR.EqualityExpression:
                OUTPUT.write(self._op_string[EXPR.EqualityExpression])
                self._print_nonlinear_terms_NL(exp.arg(0))
                self._print_nonlinear_terms_NL(exp.arg(1))

            elif isinstance(exp, (NamedExpressionData, IIdentityExpression)):
                self._print_nonlinear_terms_NL(exp.expr)

            else:
                raise ValueError(
                    "Unsupported expression type (%s) in _print_nonlinear_terms_NL"
                    % (exp_type)
                )

        elif isinstance(exp, (var.VarData, IVariable)) and (not exp.is_fixed()):
            # (self._output_fixed_variable_bounds or
            if not self._symbolic_solver_labels:
                OUTPUT.write(
                    self._op_string[var.VarData]
                    % (self.ampl_var_id[self._varID_map[id(exp)]])
                )
            else:
                OUTPUT.write(
                    self._op_string[var.VarData]
                    % (
                        self.ampl_var_id[self._varID_map[id(exp)]],
                        self._name_labeler(exp),
                    )
                )

        elif isinstance(exp, param.ParamData):
            OUTPUT.write(self._op_string[param.ParamData] % (value(exp)))

        elif isinstance(exp, NumericConstant) or exp.is_fixed():
            OUTPUT.write(self._op_string[NumericConstant] % (value(exp)))

        else:
            raise ValueError(
                "Unsupported expression type (%s) in _print_nonlinear_terms_NL"
                % (exp_type)
            )

    def _print_model_NL(
        self,
        model,
        solver_capability,
        show_section_timing=False,
        skip_trivial_constraints=False,
        file_determinism=1,
        include_all_variable_bounds=False,
        export_nonlinear_variables=False,
    ):
        output_fixed_variable_bounds = self._output_fixed_variable_bounds
        symbolic_solver_labels = self._symbolic_solver_labels

        sorter = SortComponents.unsorted
        if file_determinism >= 1:
            sorter = sorter | SortComponents.indices
            if file_determinism >= 2:
                sorter = sorter | SortComponents.alphabetical

        OUTPUT = self._OUTPUT
        assert OUTPUT is not None

        # maps NL variables to the "real" variable names in the problem.
        # it's really NL variable ordering, as there are no variable names
        # in the NL format. however, we by convention make them go from
        # x0 upward.

        overall_timer = StopWatch()
        subsection_timer = StopWatch()

        # create the symbol_map
        symbol_map = SymbolMap()

        name_labeler = self._name_labeler
        # These will get updated when symbolic_solver_labels
        # is true. It is critical that nonzero values go in
        # the header of the NL file if you want ASL to
        # parse the .row and .col files.
        max_rowname_len = 0
        max_colname_len = 0

        #
        # Collect statistics
        #
        self_ampl_var_id = self.ampl_var_id = {}
        self_ampl_con_id = self.ampl_con_id = {}
        self_ampl_obj_id = self.ampl_obj_id = {}

        # will be the entire list of all vars in the model (not
        # necessarily used)
        Vars_dict = dict()
        # will be the entire list of all objective used in the problem
        # (size=1)
        Objectives_dict = dict()
        # will be the entire list of all constraints used in the
        # problem
        Constraints_dict = dict()

        UsedVars = set()

        # linear variables
        LinearVars = set()
        LinearVarsInt = set()
        LinearVarsBool = set()

        # Tabulate the External Function definitions
        self.external_byFcn = {}
        external_Libs = set()
        for fcn in model.component_objects(ExternalFunction, active=True):
            if fcn._function in self.external_byFcn:
                if self.external_byFcn[fcn._function][0]._library != fcn._library:
                    raise RuntimeError(
                        "The same external function name (%s) is associated "
                        "with two different libraries (%s through %s, and %s "
                        "through %s).  The ASL solver will fail to link "
                        "correctly."
                        % (
                            fcn._function,
                            self.external_byFcn[fcn._function]._library,
                            self.external_byFcn[fcn._function]._library.name,
                            fcn._library,
                            fcn.name,
                        )
                    )
            else:
                self.external_byFcn[fcn._function] = (fcn, len(self.external_byFcn))
            external_Libs.add(fcn._library)
        if external_Libs:
            set_pyomo_amplfunc_env(external_Libs)
        elif "PYOMO_AMPLFUNC" in os.environ:
            del os.environ["PYOMO_AMPLFUNC"]

        subsection_timer.reset()

        # Cache the list of model blocks so we don't have to call
        # model.block_data_objects() many many times
        all_blocks_list = list(model.block_data_objects(active=True, sort=sorter))

        # create a deterministic var labeling
        Vars_dict = dict(enumerate(model.component_data_objects(Var, sort=sorter)))
        cntr = len(Vars_dict)
        # cntr = 0
        # for block in all_blocks_list:
        #     vars_counter = tuple(enumerate(
        #         block.component_data_objects(Var,
        #                                      active=True,
        #                                      sort=sorter,
        #                                      descend_into=False),
        #         cntr))
        #     cntr += len(vars_counter)
        #     Vars_dict.update(vars_counter)
        self._varID_map = dict((id(val), key) for key, val in Vars_dict.items())
        self_varID_map = self._varID_map
        # Use to label the rest of the components (which we will not encounter twice)
        trivial_labeler = _Counter(cntr)

        #
        # Count number of objectives and build the repns
        #
        n_objs = 0
        n_nonlinear_objs = 0
        ObjVars = set()
        ObjNonlinearVars = set()
        ObjNonlinearVarsInt = set()
        for block in all_blocks_list:
            gen_obj_repn = getattr(block, "_gen_obj_repn", None)
            if gen_obj_repn is not None:
                gen_obj_repn = bool(gen_obj_repn)
                # Get/Create the ComponentMap for the repn
                if not hasattr(block, '_repn'):
                    block._repn = ComponentMap()
                block_repn = block._repn

            for active_objective in block.component_data_objects(
                Objective, active=True, sort=sorter, descend_into=False
            ):
                if symbolic_solver_labels:
                    objname = name_labeler(active_objective)
                    if len(objname) > max_rowname_len:
                        max_rowname_len = len(objname)

                if gen_obj_repn == False:
                    repn = block_repn[active_objective]
                    linear_vars = repn.linear_vars
                    # By default, the NL writer generates
                    # StandardRepn objects without the more
                    # expense quadratic processing, but
                    # there is no guarantee of this if we
                    # are using a cached repn object, so we
                    # must check for the quadratic form.
                    if repn.is_nonlinear() and (repn.nonlinear_expr is None):
                        # Note that this is fragile:
                        # generate_standard_repn can leave nonlinear
                        # terms in both quadratic and nonlinear fields.
                        # However, when this was written the assumption
                        # is that generate_standard_repn is only called
                        # with quadratic=True for QCQPs (by the LP
                        # writer).  So, quadratic and nonlinear_expr
                        # will both never be non-empty.  This assertion
                        # will fail if that assumption is ever violated:
                        assert repn.is_quadratic()
                        assert len(repn.quadratic_vars) > 0
                        nonlinear_vars = {}
                        for v1, v2 in repn.quadratic_vars:
                            nonlinear_vars[id(v1)] = v1
                            nonlinear_vars[id(v2)] = v2
                        nonlinear_vars = nonlinear_vars.values()
                    else:
                        nonlinear_vars = repn.nonlinear_vars
                else:
                    repn = generate_standard_repn(
                        active_objective.expr, quadratic=False
                    )
                    linear_vars = repn.linear_vars
                    nonlinear_vars = repn.nonlinear_vars
                    if gen_obj_repn:
                        block_repn[active_objective] = repn
                try:
                    wrapped_repn = RepnWrapper(
                        repn,
                        list(self_varID_map[id(var)] for var in linear_vars),
                        list(self_varID_map[id(var)] for var in nonlinear_vars),
                    )
                except KeyError as err:
                    self._symbolMapKeyError(
                        err,
                        model,
                        self_varID_map,
                        list(linear_vars) + list(nonlinear_vars),
                    )
                    raise

                LinearVars.update(wrapped_repn.linear_vars)
                ObjNonlinearVars.update(wrapped_repn.nonlinear_vars)

                ObjVars.update(wrapped_repn.linear_vars)
                ObjVars.update(wrapped_repn.nonlinear_vars)

                obj_ID = trivial_labeler(active_objective)
                Objectives_dict[obj_ID] = (active_objective, wrapped_repn)
                self_ampl_obj_id[obj_ID] = n_objs
                symbol_map.addSymbols([(active_objective, "o%d" % n_objs)])

                n_objs += 1
                if repn.is_nonlinear():
                    n_nonlinear_objs += 1

        # I don't think this is necessarily true for the entire code base,
        # but seeing has how its never been tested we should go ahead and
        # raise an exception
        if n_objs > 1:
            raise ValueError(
                "The NL writer has detected multiple active objective functions "
                "on model %s, but currently only handles a single objective."
                % (model.name)
            )
        elif n_objs == 1:
            symbol_map.alias(symbol_map.bySymbol["o0"], "__default_objective__")

        if show_section_timing:
            subsection_timer.report("Generate objective representation")
            subsection_timer.reset()

        #
        # Count number of constraints and build the repns
        #
        n_ranges = 0
        n_single_sided_ineq = 0
        n_equals = 0
        n_unbounded = 0
        n_nonlinear_constraints = 0
        ConNonlinearVars = set()
        ConNonlinearVarsInt = set()
        nnz_grad_constraints = 0
        constraint_bounds_dict = {}
        nonlin_con_order_list = []
        lin_con_order_list = []
        ccons_lin = 0
        ccons_nonlin = 0
        ccons_nd = 0
        ccons_nzlb = 0

        for block in all_blocks_list:
            all_repns = list()

            gen_con_repn = getattr(block, "_gen_con_repn", None)
            if gen_con_repn is not None:
                gen_con_repn = bool(gen_con_repn)
                # Get/Create the ComponentMap for the repn
                if not hasattr(block, '_repn'):
                    block._repn = ComponentMap()
                block_repn = block._repn

            # Initializing the constraint dictionary
            for constraint_data in block.component_data_objects(
                Constraint, active=True, sort=sorter, descend_into=False
            ):
                if (not constraint_data.has_lb()) and (not constraint_data.has_ub()):
                    assert not constraint_data.equality
                    continue  # non-binding, so skip

                if symbolic_solver_labels:
                    conname = name_labeler(constraint_data)
                    if len(conname) > max_rowname_len:
                        max_rowname_len = len(conname)

                if gen_con_repn == False:
                    repn = block_repn[constraint_data]
                    linear_vars = repn.linear_vars
                    # By default, the NL writer generates
                    # StandardRepn objects without the more
                    # expense quadratic processing, but
                    # there is no guarantee of this if we
                    # are using a cached repn object, so we
                    # must check for the quadratic form.
                    if repn.is_nonlinear() and (repn.nonlinear_expr is None):
                        # Note that this is fragile:
                        # generate_standard_repn can leave nonlinear
                        # terms in both quadratic and nonlinear fields.
                        # However, when this was written the assumption
                        # is that generate_standard_repn is only called
                        # with quadratic=True for QCQPs (by the LP
                        # writer).  So, quadratic and nonlinear_expr
                        # will both never be non-empty.  This assertion
                        # will fail if that assumption is ever violated:
                        assert repn.is_quadratic()
                        assert len(repn.quadratic_vars) > 0
                        nonlinear_vars = {}
                        for v1, v2 in repn.quadratic_vars:
                            nonlinear_vars[id(v1)] = v1
                            nonlinear_vars[id(v2)] = v2
                        nonlinear_vars = nonlinear_vars.values()
                    else:
                        nonlinear_vars = repn.nonlinear_vars
                else:
                    if constraint_data._linear_canonical_form:
                        repn = constraint_data.canonical_form()
                        linear_vars = repn.linear_vars
                        nonlinear_vars = repn.nonlinear_vars
                    else:
                        repn = generate_standard_repn(
                            constraint_data.body, quadratic=False
                        )
                        linear_vars = repn.linear_vars
                        nonlinear_vars = repn.nonlinear_vars
                    if gen_con_repn:
                        block_repn[constraint_data] = repn

                ### GAH: Even if this is fixed, it is still useful to
                ###      write out these types of constraints
                ###      (trivial) as a feasibility check for fixed
                ###      variables, in which case the solver will pick
                ###      up on the model infeasibility.
                if skip_trivial_constraints and repn.is_fixed():
                    continue

                con_ID = trivial_labeler(constraint_data)
                try:
                    wrapped_repn = RepnWrapper(
                        repn,
                        list(self_varID_map[id(var)] for var in linear_vars),
                        list(self_varID_map[id(var)] for var in nonlinear_vars),
                    )
                except KeyError as err:
                    self._symbolMapKeyError(
                        err,
                        model,
                        self_varID_map,
                        list(linear_vars) + list(nonlinear_vars),
                    )
                    raise

                if repn.is_nonlinear():
                    nonlin_con_order_list.append(con_ID)
                    n_nonlinear_constraints += 1
                else:
                    lin_con_order_list.append(con_ID)

                Constraints_dict[con_ID] = (constraint_data, wrapped_repn)

                LinearVars.update(wrapped_repn.linear_vars)
                ConNonlinearVars.update(wrapped_repn.nonlinear_vars)

                nnz_grad_constraints += len(
                    set(wrapped_repn.linear_vars).union(wrapped_repn.nonlinear_vars)
                )

                L = None
                U = None
                if constraint_data.has_lb():
                    L = _get_bound(constraint_data.lower)
                else:
                    assert constraint_data.has_ub()
                if constraint_data.has_ub():
                    U = _get_bound(constraint_data.upper)
                else:
                    assert constraint_data.has_lb()
                if constraint_data.equality:
                    assert L == U

                offset = repn.constant
                _type = getattr(constraint_data, '_complementarity', None)
                _vid = getattr(constraint_data, '_vid', None)
                if not _type is None:
                    _vid = self_varID_map[_vid] + 1
                    constraint_bounds_dict[con_ID] = "5 {0} {1}\n".format(_type, _vid)
                    if _type == 1 or _type == 2:
                        n_single_sided_ineq += 1
                    elif _type == 3:
                        n_ranges += 1
                    elif _type == 4:
                        n_unbounded += 1
                    if repn.is_nonlinear():
                        ccons_nonlin += 1
                    else:
                        ccons_lin += 1
                else:
                    if L == U:
                        if L is None:
                            # No constraint on body
                            constraint_bounds_dict[con_ID] = "3\n"
                            n_unbounded += 1
                        else:
                            constraint_bounds_dict[con_ID] = "4 %r\n" % (L - offset)
                            n_equals += 1
                    elif L is None:
                        constraint_bounds_dict[con_ID] = "1 %r\n" % (U - offset)
                        n_single_sided_ineq += 1
                    elif U is None:
                        constraint_bounds_dict[con_ID] = "2 %r\n" % (L - offset)
                        n_single_sided_ineq += 1
                    elif L > U:
                        msg = (
                            'Constraint {0}: lower bound greater than upper'
                            ' bound ({1} > {2})'
                        )
                        raise ValueError(
                            msg.format(constraint_data.name, str(L), str(U))
                        )
                    else:
                        constraint_bounds_dict[con_ID] = "0 %r %r\n" % (
                            L - offset,
                            U - offset,
                        )
                        # double sided inequality
                        # both are not none and they are valid
                        n_ranges += 1

        sos1 = solver_capability("sos1")
        sos2 = solver_capability("sos2")
        for block in all_blocks_list:
            for soscondata in block.component_data_objects(
                SOSConstraint, active=True, sort=sorter, descend_into=False
            ):
                level = soscondata.level
                if (level == 1 and not sos1) or (level == 2 and not sos2):
                    raise Exception(
                        "Solver does not support SOS level %s constraints" % (level,)
                    )
                if hasattr(soscondata, "get_variables"):
                    LinearVars.update(
                        self_varID_map[id(vardata)]
                        for vardata in soscondata.get_variables()
                    )
                else:
                    LinearVars.update(
                        self_varID_map[id(vardata)] for vardata in soscondata.variables
                    )

        # create the ampl constraint ids
        self_ampl_con_id.update(
            (con_ID, row_id)
            for row_id, con_ID in enumerate(
                itertools.chain(nonlin_con_order_list, lin_con_order_list)
            )
        )
        # populate the symbol_map
        symbol_map.addSymbols(
            [
                (Constraints_dict[con_ID][0], "c%d" % row_id)
                for row_id, con_ID in enumerate(
                    itertools.chain(nonlin_con_order_list, lin_con_order_list)
                )
            ]
        )

        if show_section_timing:
            subsection_timer.report("Generate constraint representations")
            subsection_timer.reset()

        UsedVars.update(LinearVars)
        UsedVars.update(ObjNonlinearVars)
        UsedVars.update(ConNonlinearVars)

        LinearVars = LinearVars.difference(ObjNonlinearVars)
        LinearVars = LinearVars.difference(ConNonlinearVars)

        if include_all_variable_bounds:
            # classify unused vars as linear
            AllVars = set(self_varID_map[id(vardata)] for vardata in Vars_dict.values())
            UnusedVars = AllVars.difference(UsedVars)
            LinearVars.update(UnusedVars)

        if export_nonlinear_variables:
            for v in export_nonlinear_variables:
                v_iter = v.values() if v.is_indexed() else iter((v,))
                for vi in v_iter:
                    if self_varID_map[id(vi)] not in UsedVars:
                        Vars_dict[id(vi)] = vi
                        ConNonlinearVars.update([self_varID_map[id(vi)]])

        ### There used to be an if statement here for the following code block
        ### checking model.statistics.num_binary_vars was greater than zero.
        ### To this day, I don't know how it worked.
        ### TODO: Figure out why
        ###############

        for var_ID in LinearVars:
            var = Vars_dict[var_ID]
            if var.is_binary():
                L = var.lb
                U = var.ub
                if (L is None) or (U is None):
                    raise ValueError(
                        "Variable "
                        + str(var.name)
                        + "is binary, but does not have lb and ub set"
                    )
                LinearVarsBool.add(var_ID)
            elif var.is_integer():
                LinearVarsInt.add(var_ID)
            elif not var.is_continuous():
                raise TypeError(
                    "Invalid domain type for variable with name '%s'. "
                    "Variable is not continuous, integer, or binary."
                )
        LinearVars.difference_update(LinearVarsInt)
        LinearVars.difference_update(LinearVarsBool)

        for var_ID in ObjNonlinearVars:
            var = Vars_dict[var_ID]
            if var.is_integer() or var.is_binary():
                ObjNonlinearVarsInt.add(var_ID)
            elif not var.is_continuous():
                raise TypeError(
                    "Invalid domain type for variable with name '%s'. "
                    "Variable is not continuous, integer, or binary."
                )
        ObjNonlinearVars.difference_update(ObjNonlinearVarsInt)
        for var_ID in ConNonlinearVars:
            var = Vars_dict[var_ID]
            if var.is_integer() or var.is_binary():
                ConNonlinearVarsInt.add(var_ID)
            elif not var.is_continuous():
                raise TypeError(
                    "Invalid domain type for variable with name '%s'. "
                    "Variable is not continuous, integer, or binary."
                )
        ConNonlinearVars.difference_update(ConNonlinearVarsInt)
        ##################

        Nonlinear_Vars_in_Objs_and_Constraints = ObjNonlinearVars.intersection(
            ConNonlinearVars
        )
        Discrete_Nonlinear_Vars_in_Objs_and_Constraints = (
            ObjNonlinearVarsInt.intersection(ConNonlinearVarsInt)
        )
        ObjNonlinearVars = ObjNonlinearVars.difference(
            Nonlinear_Vars_in_Objs_and_Constraints
        )
        ConNonlinearVars = ConNonlinearVars.difference(
            Nonlinear_Vars_in_Objs_and_Constraints
        )
        ObjNonlinearVarsInt = ObjNonlinearVarsInt.difference(
            Discrete_Nonlinear_Vars_in_Objs_and_Constraints
        )
        ConNonlinearVarsInt = ConNonlinearVarsInt.difference(
            Discrete_Nonlinear_Vars_in_Objs_and_Constraints
        )

        # put the ampl variable id into the variable
        full_var_list = []
        full_var_list.extend(sorted(Nonlinear_Vars_in_Objs_and_Constraints))
        full_var_list.extend(sorted(Discrete_Nonlinear_Vars_in_Objs_and_Constraints))
        idx_nl_both = len(full_var_list)
        #
        full_var_list.extend(sorted(ConNonlinearVars))
        full_var_list.extend(sorted(ConNonlinearVarsInt))
        idx_nl_con = len(full_var_list)
        #
        full_var_list.extend(sorted(ObjNonlinearVars))
        full_var_list.extend(sorted(ObjNonlinearVarsInt))
        idx_nl_obj = len(full_var_list)
        #
        full_var_list.extend(sorted(LinearVars))
        full_var_list.extend(sorted(LinearVarsBool))
        full_var_list.extend(sorted(LinearVarsInt))

        if idx_nl_obj == idx_nl_con:
            idx_nl_obj = idx_nl_both

        # create the ampl variable column ids
        self_ampl_var_id.update(
            (var_ID, column_id) for column_id, var_ID in enumerate(full_var_list)
        )
        # populate the symbol_map
        symbol_map.addSymbols(
            [
                (Vars_dict[var_ID], "v%d" % column_id)
                for column_id, var_ID in enumerate(full_var_list)
            ]
        )

        if show_section_timing:
            subsection_timer.report("Partition variable types")
            subsection_timer.reset()

        #        end_time = time.clock()
        #        print (end_time - start_time)

        colfilename = None
        if OUTPUT.name.endswith('.nl'):
            colfilename = OUTPUT.name.replace('.nl', '.col')
        else:
            colfilename = OUTPUT.name + '.col'
        if symbolic_solver_labels:
            colf = open(colfilename, 'w')
            colfile_line_template = "%s\n"
            for var_ID in full_var_list:
                varname = name_labeler(Vars_dict[var_ID])
                colf.write(colfile_line_template % varname)
                if len(varname) > max_colname_len:
                    max_colname_len = len(varname)
            colf.close()

        if show_section_timing:
            subsection_timer.report("Write .col file")
            subsection_timer.reset()

        if len(full_var_list) < 1:
            raise ValueError(
                "No variables appear in the Pyomo model constraints or"
                " objective. This is not supported by the NL file interface"
            )

        #
        # Print Header
        #
        # LINE 1
        #
        OUTPUT.write("g3 1 1 0\t# problem {0}\n".format(model.name))
        #
        # LINE 2
        #
        OUTPUT.write(
            " {0} {1} {2} {3} {4} \t# vars, constraints, "
            "objectives, ranges, eqns\n".format(
                len(full_var_list),
                n_single_sided_ineq + n_ranges + n_equals + n_unbounded,
                n_objs,
                n_ranges,
                n_equals,
            )
        )
        #
        # LINE 3
        #
        OUTPUT.write(
            " {0} {1} {2} {3} {4} {5}\t# nonlinear constrs, "
            "objs; ccons: lin, nonlin, nd, nzlb\n".format(
                n_nonlinear_constraints,
                n_nonlinear_objs,
                ccons_lin,
                ccons_nonlin,
                ccons_nd,
                ccons_nzlb,
            )
        )
        #
        # LINE 4
        #
        OUTPUT.write(" 0 0\t# network constraints: nonlinear, linear\n")
        #
        # LINE 5
        #
        OUTPUT.write(
            " {0} {1} {2} \t# nonlinear vars in constraints, "
            "objectives, both\n".format(idx_nl_con, idx_nl_obj, idx_nl_both)
        )

        #
        # LINE 6
        #
        OUTPUT.write(
            " 0 {0} 0 1\t# linear network variables; functions; "
            "arith, flags\n".format(len(self.external_byFcn))
        )
        #
        # LINE 7
        #
        n_int_nonlinear_b = len(Discrete_Nonlinear_Vars_in_Objs_and_Constraints)
        n_int_nonlinear_c = len(ConNonlinearVarsInt)
        n_int_nonlinear_o = len(ObjNonlinearVarsInt)
        OUTPUT.write(
            " {0} {1} {2} {3} {4} \t# discrete variables: binary, "
            "integer, nonlinear (b,c,o)\n".format(
                len(LinearVarsBool),
                len(LinearVarsInt),
                n_int_nonlinear_b,
                n_int_nonlinear_c,
                n_int_nonlinear_o,
            )
        )
        #
        # LINE 8
        #
        # objective info computed above
        OUTPUT.write(
            " {0} {1} \t# nonzeros in Jacobian, obj. gradient\n".format(
                nnz_grad_constraints, len(ObjVars)
            )
        )
        #
        # LINE 9
        #
        OUTPUT.write(
            " %d %d\t# max name lengths: constraints, variables\n"
            % (max_rowname_len, max_colname_len)
        )

        #
        # LINE 10
        #
        OUTPUT.write(" 0 0 0 0 0\t# common exprs: b,c,o,c1,o1\n")

        #        end_time = time.clock()
        #        print (end_time - start_time)

        #        print "Printing constraints:",
        #        start_time = time.clock()

        #
        # "F" lines
        #
        for fcn, fid in sorted(
            self.external_byFcn.values(), key=operator.itemgetter(1)
        ):
            OUTPUT.write("F%d 1 -1 %s\n" % (fid, fcn._function))

        #
        # "S" lines
        #

        # Translate the SOSConstraint component into ampl suffixes
        sos1 = solver_capability("sos1")
        sos2 = solver_capability("sos2")
        modelSOS = ModelSOS(self_ampl_var_id, self_varID_map)
        for block in all_blocks_list:
            for soscondata in block.component_data_objects(
                SOSConstraint, active=True, sort=sorter, descend_into=False
            ):
                level = soscondata.level
                if (level == 1 and not sos1) or (level == 2 and not sos2):
                    raise ValueError(
                        "Solver does not support SOS level %s constraints" % (level)
                    )
                modelSOS.count_constraint(soscondata)

        symbol_map_byObject = symbol_map.byObject

        var_sosno_suffix = modelSOS.sosno
        var_ref_suffix = modelSOS.ref
        sosconstraint_sosno_vals = set(var_sosno_suffix.vals)

        # Translate the rest of the Pyomo Suffix components
        suffix_header_line = "S{0} {1} {2}\n"
        suffix_line = "{0} {1!r}\n"
        var_tag = 0
        con_tag = 1
        obj_tag = 2
        prob_tag = 3
        suffix_dict = {}
        if isinstance(model, IBlock):
            suffix_gen = lambda b: (
                (suf.storage_key, suf)
                for suf in pyomo.core.kernel.suffix.export_suffix_generator(
                    b, active=True, descend_into=False
                )
            )
        else:
            suffix_gen = (
                lambda b: pyomo.core.base.suffix.active_export_suffix_generator(b)
            )
        for block in all_blocks_list:
            for name, suf in suffix_gen(block):
                if len(suf):
                    suffix_dict.setdefault(name, []).append(suf)
        if not ('sosno' in suffix_dict):
            # We still need to write out the SOSConstraint suffixes
            # even though these may have not been "declared" on the model
            s_lines = var_sosno_suffix.genfilelines()
            len_s_lines = len(s_lines)
            if len_s_lines > 0:
                OUTPUT.write(suffix_header_line.format(var_tag, len_s_lines, 'sosno'))
                OUTPUT.writelines(s_lines)
        else:
            # I am choosing not to allow a user to mix the use of the Pyomo
            # SOSConstraint component and manual sosno declarations within
            # a single model. I initially tried to allow this but the
            # var section of the code below blows up for two reason. (1)
            # we have to make sure that the sosno suffix is not defined
            # twice for the same variable (2) We have to make sure that
            # the automatically chosen sosno used by the SOSConstraint
            # translation does not already match one a user has manually
            # implemented (this would modify the members in an sos set).
            # Since this suffix is exclusively used for defining sos sets,
            # there is no reason a user can not just stick to one method.
            if not var_sosno_suffix.is_empty():
                raise RuntimeError(
                    "The Pyomo NL file writer does not allow both manually "
                    "declared 'sosno' suffixes as well as SOSConstraint "
                    "components to exist on a single model. To avoid this "
                    "error please use only one of these methods to define "
                    "special ordered sets."
                )
        if not ('ref' in suffix_dict):
            # We still need to write out the SOSConstraint suffixes
            # even though these may have not been "declared" on the model
            s_lines = var_ref_suffix.genfilelines()
            len_s_lines = len(s_lines)
            if len_s_lines > 0:
                OUTPUT.write(suffix_header_line.format(var_tag, len_s_lines, 'ref'))
                OUTPUT.writelines(s_lines)
        else:
            # see reason (1) in the paragraph above for why we raise this
            # exception (replacing sosno with ref).
            if not var_ref_suffix.is_empty():
                raise RuntimeError(
                    "The Pyomo NL file writer does not allow both manually "
                    "declared 'ref' suffixes as well as SOSConstraint "
                    "components to exist on a single model. To avoid this "
                    "error please use only one of these methods to define "
                    "special ordered sets."
                )
        # do a sort to make sure NL file output is deterministic
        # across python versions
        for suffix_name in sorted(suffix_dict):
            suffixes = suffix_dict[suffix_name]
            datatypes = set()
            for suffix in suffixes:
                try:
                    datatype = suffix.datatype
                except AttributeError:
                    datatype = suffix.get_datatype()
                if datatype not in (Suffix.FLOAT, Suffix.INT):
                    raise ValueError(
                        "The Pyomo NL file writer requires that all active export "
                        "Suffix components declare a numeric datatype. Suffix "
                        "component: %s with " % (suffix_name)
                    )
                datatypes.add(datatype)
            if len(datatypes) != 1:
                raise ValueError(
                    "The Pyomo NL file writer found multiple active export suffix "
                    "components with name %s with different datatypes. A single "
                    "datatype must be declared." % (suffix_name)
                )
            if suffix_name == "dual":
                # The NL file format has a special section for dual initializations
                continue
            float_tag = 0
            if datatypes.pop() == Suffix.FLOAT:
                float_tag = 4

            var_s_lines = []
            con_s_lines = []
            obj_s_lines = []
            mod_s_lines = []
            for suffix in suffixes:
                for component_data, suffix_value in suffix.items():
                    try:
                        symbol = symbol_map_byObject[id(component_data)]
                        type_tag = symbol[0]
                        ampl_id = int(symbol[1:])
                        if type_tag == 'v':
                            var_s_lines.append((ampl_id, suffix_value))
                        elif type_tag == 'c':
                            con_s_lines.append((ampl_id, suffix_value))
                        elif type_tag == 'o':
                            obj_s_lines.append((ampl_id, suffix_value))
                        else:
                            # This would be a developer error
                            assert False
                    except KeyError:
                        if component_data is model:
                            mod_s_lines.append((0, suffix_value))

            ################## vars
            if len(var_s_lines) > 0:
                OUTPUT.write(
                    suffix_header_line.format(
                        var_tag | float_tag, len(var_s_lines), suffix_name
                    )
                )
                OUTPUT.writelines(
                    suffix_line.format(*_l)
                    for _l in sorted(var_s_lines, key=operator.itemgetter(0))
                )
            ################## constraints
            if len(con_s_lines) > 0:
                OUTPUT.write(
                    suffix_header_line.format(
                        con_tag | float_tag, len(con_s_lines), suffix_name
                    )
                )
                OUTPUT.writelines(
                    suffix_line.format(*_l)
                    for _l in sorted(con_s_lines, key=operator.itemgetter(0))
                )
            ################## objectives
            if len(obj_s_lines) > 0:
                OUTPUT.write(
                    suffix_header_line.format(
                        obj_tag | float_tag, len(obj_s_lines), suffix_name
                    )
                )
                OUTPUT.writelines(
                    suffix_line.format(*_l)
                    for _l in sorted(obj_s_lines, key=operator.itemgetter(0))
                )
            ################## problems (in this case the one problem)
            if len(mod_s_lines) > 0:
                if len(mod_s_lines) > 1:
                    logger.warning(
                        "ProblemWriter_nl: Collected multiple values for Suffix %s "
                        "referencing model %s. This is likely a bug."
                        % (suffix_name, model.name)
                    )
                OUTPUT.write(
                    suffix_header_line.format(
                        prob_tag | float_tag, len(mod_s_lines), suffix_name
                    )
                )
                OUTPUT.writelines(
                    suffix_line.format(*_l)
                    for _l in sorted(mod_s_lines, key=operator.itemgetter(0))
                )

        del modelSOS

        #
        # "C" lines
        #
        rowfilename = None
        if OUTPUT.name.endswith('.nl'):
            rowfilename = OUTPUT.name.replace('.nl', '.row')
        else:
            rowfilename = OUTPUT.name + '.row'
        if symbolic_solver_labels:
            rowf = open(rowfilename, 'w')

        cu = [0 for i in range(len(full_var_list))]
        for con_ID in nonlin_con_order_list:
            con_data, wrapped_repn = Constraints_dict[con_ID]
            row_id = self_ampl_con_id[con_ID]
            OUTPUT.write("C%d" % (row_id))
            if symbolic_solver_labels:
                lbl = name_labeler(con_data)
                OUTPUT.write("\t#%s" % (lbl))
                rowf.write(lbl + "\n")
            OUTPUT.write("\n")

            if wrapped_repn.repn.nonlinear_expr is not None:
                assert not wrapped_repn.repn.is_quadratic()
                self._print_nonlinear_terms_NL(wrapped_repn.repn.nonlinear_expr)
            else:
                assert wrapped_repn.repn.is_quadratic()
                self._print_standard_quadratic_NL(
                    wrapped_repn.repn.quadratic_vars, wrapped_repn.repn.quadratic_coefs
                )

            for var_ID in set(wrapped_repn.linear_vars).union(
                wrapped_repn.nonlinear_vars
            ):
                cu[self_ampl_var_id[var_ID]] += 1

        for con_ID in lin_con_order_list:
            con_data, wrapped_repn = Constraints_dict[con_ID]
            row_id = self_ampl_con_id[con_ID]
            con_vars = set(wrapped_repn.linear_vars)
            for var_ID in con_vars:
                cu[self_ampl_var_id[var_ID]] += 1
            OUTPUT.write("C%d" % (row_id))
            if symbolic_solver_labels:
                lbl = name_labeler(con_data)
                OUTPUT.write("\t#%s" % (lbl))
                rowf.write(lbl + "\n")
            OUTPUT.write("\n")
            OUTPUT.write("n0\n")

        if show_section_timing:
            subsection_timer.report("Write NL header and suffix lines")
            subsection_timer.reset()

        #
        # "O" lines
        #
        for obj_ID, (obj, wrapped_repn) in Objectives_dict.items():
            k = 0
            if not obj.is_minimizing():
                k = 1

            OUTPUT.write("O%d %d" % (self_ampl_obj_id[obj_ID], k))
            if symbolic_solver_labels:
                lbl = name_labeler(obj)
                OUTPUT.write("\t#%s" % (lbl))
                rowf.write(lbl + "\n")
            OUTPUT.write("\n")

            if wrapped_repn.repn.is_linear():
                OUTPUT.write(
                    self._op_string[NumericConstant] % (wrapped_repn.repn.constant)
                )
            else:
                if wrapped_repn.repn.constant != 0:
                    _, binary_sum_str, _ = self._op_string[EXPR.SumExpressionBase]
                    OUTPUT.write(binary_sum_str)
                    OUTPUT.write(
                        self._op_string[NumericConstant] % (wrapped_repn.repn.constant)
                    )
                if wrapped_repn.repn.nonlinear_expr is not None:
                    assert not wrapped_repn.repn.is_quadratic()
                    self._print_nonlinear_terms_NL(wrapped_repn.repn.nonlinear_expr)
                else:
                    assert wrapped_repn.repn.is_quadratic()
                    self._print_standard_quadratic_NL(
                        wrapped_repn.repn.quadratic_vars,
                        wrapped_repn.repn.quadratic_coefs,
                    )

        if symbolic_solver_labels:
            rowf.close()
        del name_labeler

        if show_section_timing:
            subsection_timer.report("Write objective expression")
            subsection_timer.reset()

        #
        # "d" lines
        #
        # dual initialization
        if 'dual' in suffix_dict:
            s_lines = []
            for dual_suffix in suffix_dict['dual']:
                for constraint_data, suffix_value in dual_suffix.items():
                    try:
                        # a constraint might not be referenced
                        # (inactive / on inactive block)
                        symbol = symbol_map_byObject[id(constraint_data)]
                        type_tag = symbol[0]
                        assert type_tag == 'c'
                        ampl_con_id = int(symbol[1:])
                        s_lines.append((ampl_con_id, suffix_value))
                    except KeyError:
                        pass

            if len(s_lines) > 0:
                OUTPUT.write("d%d" % (len(s_lines)))
                if symbolic_solver_labels:
                    OUTPUT.write("\t# dual initial guess")
                OUTPUT.write("\n")
                OUTPUT.writelines(
                    suffix_line.format(*_l)
                    for _l in sorted(s_lines, key=operator.itemgetter(0))
                )

        #
        # "x" lines
        #
        # variable initialization
        var_bound_list = []
        x_init_list = []
        for ampl_var_id, var_ID in enumerate(full_var_list):
            var = Vars_dict[var_ID]
            if var.value is not None:
                x_init_list.append("%d %r\n" % (ampl_var_id, var.value))
            if var.fixed:
                if not output_fixed_variable_bounds:
                    raise ValueError(
                        "Encountered a fixed variable (%s) inside an active objective"
                        " or constraint expression on model %s, which is usually "
                        "indicative of a preprocessing error. Use the IO-option "
                        "'output_fixed_variable_bounds=True' to suppress this error "
                        "and fix the variable by overwriting its bounds in the NL "
                        "file." % (var.name, model.name)
                    )
                if var.value is None:
                    raise ValueError("Variable cannot be fixed to a value of None.")
                L = U = _get_bound(var.value)
            else:
                L = None
                if var.has_lb():
                    L = _get_bound(var.lb)
                U = None
                if var.has_ub():
                    U = _get_bound(var.ub)
            if L is not None:
                if U is not None:
                    if L == U:
                        var_bound_list.append("4 %r\n" % (L))
                    else:
                        var_bound_list.append("0 %r %r\n" % (L, U))
                else:
                    var_bound_list.append("2 %r\n" % (L))
            elif U is not None:
                var_bound_list.append("1 %r\n" % (U))
            else:
                var_bound_list.append("3\n")

        OUTPUT.write("x%d" % (len(x_init_list)))
        if symbolic_solver_labels:
            OUTPUT.write("\t# initial guess")
        OUTPUT.write("\n")
        OUTPUT.writelines(x_init_list)
        del x_init_list

        if show_section_timing:
            subsection_timer.report("Write initializations")
            subsection_timer.reset()

        #
        # "r" lines
        #
        OUTPUT.write("r")
        if symbolic_solver_labels:
            OUTPUT.write(
                "\t#%d ranges (rhs's)"
                % (len(nonlin_con_order_list) + len(lin_con_order_list))
            )
        OUTPUT.write("\n")
        # *NOTE: This iteration follows the assignment of the ampl_con_id
        OUTPUT.writelines(
            constraint_bounds_dict[con_ID]
            for con_ID in itertools.chain(nonlin_con_order_list, lin_con_order_list)
        )

        if show_section_timing:
            subsection_timer.report("Write constraint bounds")
            subsection_timer.reset()

        #
        # "b" lines
        #
        OUTPUT.write("b")
        if symbolic_solver_labels:
            OUTPUT.write("\t#%d bounds (on variables)" % (len(var_bound_list)))
        OUTPUT.write("\n")
        OUTPUT.writelines(var_bound_list)
        del var_bound_list

        if show_section_timing:
            subsection_timer.report("Write variable bounds")
            subsection_timer.reset()

        #
        # "k" lines
        #
        ktot = 0
        n1 = len(full_var_list) - 1
        OUTPUT.write("k%d" % (n1))
        if symbolic_solver_labels:
            OUTPUT.write("\t#intermediate Jacobian column lengths")
        OUTPUT.write("\n")
        ktot = 0
        for i in range(n1):
            ktot += cu[i]
            OUTPUT.write("%d\n" % (ktot))
        del cu

        if show_section_timing:
            subsection_timer.report("Write k lines")
            subsection_timer.reset()

        #
        # "J" lines
        #
        for nc, con_ID in enumerate(
            itertools.chain(nonlin_con_order_list, lin_con_order_list)
        ):
            con_data, wrapped_repn = Constraints_dict[con_ID]
            numnonlinear_vars = len(wrapped_repn.nonlinear_vars)
            numlinear_vars = len(wrapped_repn.linear_vars)
            if numnonlinear_vars == 0:
                if numlinear_vars > 0:
                    linear_dict = dict(
                        (var_ID, coef)
                        for var_ID, coef in zip(
                            wrapped_repn.linear_vars, wrapped_repn.repn.linear_coefs
                        )
                    )
                    OUTPUT.write("J%d %d\n" % (nc, numlinear_vars))
                    OUTPUT.writelines(
                        "%d %r\n" % (self_ampl_var_id[con_var], linear_dict[con_var])
                        for con_var in sorted(linear_dict.keys())
                    )
            elif numlinear_vars == 0:
                nl_con_vars = sorted(wrapped_repn.nonlinear_vars)
                OUTPUT.write("J%d %d\n" % (nc, numnonlinear_vars))
                OUTPUT.writelines(
                    "%d 0\n" % (self_ampl_var_id[con_var]) for con_var in nl_con_vars
                )
            else:
                con_vars = set(wrapped_repn.nonlinear_vars)
                nl_con_vars = sorted(con_vars.difference(wrapped_repn.linear_vars))
                con_vars.update(wrapped_repn.linear_vars)
                linear_dict = dict(
                    (var_ID, coef)
                    for var_ID, coef in zip(
                        wrapped_repn.linear_vars, wrapped_repn.repn.linear_coefs
                    )
                )
                OUTPUT.write("J%d %d\n" % (nc, len(con_vars)))
                OUTPUT.writelines(
                    "%d %r\n" % (self_ampl_var_id[con_var], linear_dict[con_var])
                    for con_var in sorted(linear_dict.keys())
                )
                OUTPUT.writelines(
                    "%d 0\n" % (self_ampl_var_id[con_var]) for con_var in nl_con_vars
                )

        if show_section_timing:
            subsection_timer.report("Write J lines")
            subsection_timer.reset()

        #
        # "G" lines
        #
        for obj_ID, (obj, wrapped_repn) in Objectives_dict.items():
            grad_entries = {}
            for idx, obj_var in enumerate(wrapped_repn.linear_vars):
                grad_entries[self_ampl_var_id[obj_var]] = (
                    wrapped_repn.repn.linear_coefs[idx]
                )
            for obj_var in wrapped_repn.nonlinear_vars:
                if obj_var not in wrapped_repn.linear_vars:
                    grad_entries[self_ampl_var_id[obj_var]] = 0
            len_ge = len(grad_entries)
            if len_ge > 0:
                OUTPUT.write("G%d %d\n" % (self_ampl_obj_id[obj_ID], len_ge))
                for var_ID in sorted(grad_entries.keys()):
                    OUTPUT.write("%d %r\n" % (var_ID, grad_entries[var_ID]))

        if show_section_timing:
            subsection_timer.report("Write G lines")
            subsection_timer.reset()
            overall_timer.report("Total time")

        return symbol_map

    def _symbolMapKeyError(self, err, model, map, vars):
        _errors = []
        for v in vars:
            if id(v) in map:
                continue
            if v.model() is not model.model():
                _errors.append(
                    "Variable '%s' is not part of the model "
                    "being written out, but appears in an "
                    "expression used on this model." % (v.name,)
                )
            else:
                _parent = v.parent_block()
                while _parent is not None and _parent is not model:
                    if _parent.ctype is not model.ctype:
                        _errors.append(
                            "Variable '%s' exists within %s '%s', "
                            "but is used by an active "
                            "expression.  Currently variables "
                            "must be reachable through a tree "
                            "of active Blocks."
                            % (v.name, _parent.ctype.__name__, _parent.name)
                        )
                    if not _parent.active:
                        _errors.append(
                            "Variable '%s' exists within "
                            "deactivated %s '%s', but is used by "
                            "an active expression.  Currently "
                            "variables must be reachable through "
                            "a tree of active Blocks."
                            % (v.name, _parent.ctype.__name__, _parent.name)
                        )
                    _parent = _parent.parent_block()

        if _errors:
            for e in _errors:
                logger.error(e)
            err.args = err.args + tuple(_errors)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import re

from difflib import SequenceMatcher, unified_diff

from pyomo.repn.tests.diffutils import compare_floats, load_baseline

_strip_comment = re.compile(r'\s*\\.*')


def _update_subsets(subset, base, test):
    for i, j in zip(*subset):
        if compare_floats(base[i], test[j]):
            base[i] = test[j]


def _preprocess_data(data):
    for line in data.splitlines():
        fields = line.split()
        for i, f in enumerate(fields):
            try:
                if int(f) == float(f):
                    fields[i] = str(int(f))
            except:
                pass
        yield ' '.join(fields)


def lp_diff(base, test, baseline='baseline', testfile='testfile'):
    if test == base:
        return [], []

    test = list(_preprocess_data(test))
    base = list(_preprocess_data(base))
    if test == base:
        return [], []

    for group in SequenceMatcher(None, base, test).get_grouped_opcodes(3):
        for tag, i1, i2, j1, j2 in group:
            if tag != 'replace':
                continue
            _update_subsets((range(i1, i2), range(j1, j2)), base, test)

    if test == base:
        return [], []

    print(
        ''.join(
            unified_diff(
                [_ + "\n" for _ in base],
                [_ + "\n" for _ in test],
                fromfile=baseline,
                tofile=testfile,
            )
        )
    )
    return base, test


def load_lp_baseline(baseline, testfile, version='lp'):
    return load_baseline(baseline, testfile, 'lp', version)


def load_and_compare_lp_baseline(baseline, testfile, version='lp'):
    return lp_diff(*load_lp_baseline(baseline, testfile, version))


if __name__ == '__main__':
    import sys

    base, test = load_and_compare_lp_baseline(sys.argv[1], sys.argv[2])
    sys.exit(1 if base or test else 0)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
import copy
import itertools

from pyomo.core import Block, ConstraintList, Set, Constraint
from pyomo.core.base import Reference
from pyomo.common.modeling import unique_component_name
from pyomo.gdp.disjunct import Disjunct, Disjunction

import logging

logger = logging.getLogger('pyomo.gdp')


def _clone_all_but_indicator_vars(self):
    """Clone everything in a Disjunct except for the indicator_vars"""
    return self.clone(
        {
            id(self.indicator_var): self.indicator_var,
            id(self.binary_indicator_var): self.binary_indicator_var,
        }
    )


def _squish_singletons(tuple_iter):
    """Squish all singleton tuples into their non-tuple values."""
    for tup in tuple_iter:
        if len(tup) == 1:
            yield tup[0]
        else:
            yield tup


def apply_basic_step(disjunctions_or_constraints):
    #
    # Basic steps only apply to XOR'd disjunctions
    #
    disjunctions = list(
        obj for obj in disjunctions_or_constraints if obj.ctype is Disjunction
    )
    constraints = list(
        obj for obj in disjunctions_or_constraints if obj.ctype is Constraint
    )
    if len(disjunctions) + len(constraints) != len(disjunctions_or_constraints):
        raise ValueError(
            'apply_basic_step only accepts a list containing '
            'Disjunctions or Constraints'
        )
    if not disjunctions:
        raise ValueError(
            'apply_basic_step: argument list must contain at least one Disjunction'
        )
    for d in disjunctions:
        if not d.xor:
            raise ValueError(
                "Basic steps can only be applied to XOR'd disjunctions\n\t"
                "(raised by disjunction %s)" % (d.name,)
            )
        if not d.active:
            logger.warning(
                "Warning: applying basic step to a previously "
                "deactivated disjunction (%s)" % (d.name,)
            )

    ans = Block(concrete=True)
    ans.DISJUNCTIONS = Set(initialize=range(len(disjunctions)))
    ans.INDEX = Set(
        dimen=len(disjunctions),
        initialize=_squish_singletons(
            itertools.product(*tuple(range(len(d.disjuncts)) for d in disjunctions))
        ),
    )

    #
    # Form the individual disjuncts for the new basic step
    #
    ans.disjuncts = Disjunct(ans.INDEX)
    for idx in ans.INDEX:
        #
        # Each source disjunct will be copied (cloned) into its own
        # subblock
        #
        ans.disjuncts[idx].src = Block(ans.DISJUNCTIONS)
        for i in ans.DISJUNCTIONS:
            src_disj = disjunctions[i].disjuncts[
                idx[i] if isinstance(idx, tuple) else idx
            ]
            tmp = _clone_all_but_indicator_vars(src_disj)
            for k, v in list(tmp.component_map().items()):
                if v.parent_block() is not tmp:
                    # Skip indicator_var and binary_indicator_var
                    continue
                tmp.del_component(k)
                ans.disjuncts[idx].src[i].add_component(k, v)
        # Copy in the constraints corresponding to the improper disjunctions
        ans.disjuncts[idx].improper_constraints = ConstraintList()
        for constr in constraints:
            if constr.is_indexed():
                for j in constr:
                    ans.disjuncts[idx].improper_constraints.add(
                        (constr[j].lower, constr[j].body, constr[j].upper)
                    )
                    constr[j].deactivate()
            # need this so that we can take an improper basic step with a
            # ConstraintData
            else:
                ans.disjuncts[idx].improper_constraints.add(
                    (constr.lower, constr.body, constr.upper)
                )
                constr.deactivate()

    #
    # Link the new disjunct indicator_var's to the original
    # indicator_var's.  Since only one of the new
    #
    ans.indicator_links = ConstraintList()
    for i in ans.DISJUNCTIONS:
        for j in range(len(disjunctions[i].disjuncts)):
            orig_var = disjunctions[i].disjuncts[j].indicator_var
            orig_binary_var = orig_var.get_associated_binary()
            ans.indicator_links.add(
                orig_binary_var
                == sum(
                    ans.disjuncts[idx].binary_indicator_var
                    for idx in ans.INDEX
                    if (idx[i] if isinstance(idx, tuple) else idx) == j
                )
            )
            # and throw on a Reference to original on the block
            for v in (orig_var, orig_binary_var):
                name_base = v.getname(fully_qualified=True)
                ans.add_component(unique_component_name(ans, name_base), Reference(v))

    # Form the new disjunction
    ans.disjunction = Disjunction(expr=[ans.disjuncts[i] for i in ans.INDEX])

    #
    # Deactivate the old disjunctions / disjuncts
    #
    for i in ans.DISJUNCTIONS:
        disjunctions[i].deactivate()
        for d in disjunctions[i].disjuncts:
            d._deactivate_without_fixing_indicator()

    return ans


if __name__ == '__main__':
    from pyomo.environ import ConcreteModel, Constraint, Var

    m = ConcreteModel()

    def _d(d, i):
        d.x = Var(range(i))
        d.silly = Constraint(expr=d.indicator_var == i)

    m.d = Disjunct([1, 2], rule=_d)

    def _e(e, i):
        e.y = Var(range(2, i))

    m.e = Disjunct([3, 4, 5], rule=_e)

    m.dd = Disjunction(expr=[m.d[1], m.d[2]])
    m.ee = Disjunction(expr=[m.e[3], m.e[4], m.e[5]])
    m.Z = apply_basic_step([m.dd, m.ee])

    m.pprint()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from itertools import product
from math import ceil, floor
from collections import Counter

from pyomo.common.dependencies import numpy as np

import pyomo.environ as pyo

"""
This script has collection of test cases that can be used to enumerate solutions.
That is, simple problems where the alternative solutions can be found manually.
"""


def _is_satisfied(constraint, feasibility_tol=1e-6):
    value = pyo.value(constraint.body)
    if constraint.has_lb() and value < constraint.lb - feasibility_tol:
        return False
    if constraint.has_ub() and value > constraint.ub + feasibility_tol:
        return False
    return True


def get_2d_diamond_problem(discrete_x=False, discrete_y=False):
    """
    Simple 2d problem where the feasible is diamond-shaped.
    """
    m = pyo.ConcreteModel()
    m.x = pyo.Var(within=pyo.Integers if discrete_x else pyo.Reals, bounds=(-10, 10))
    m.y = pyo.Var(within=pyo.Integers if discrete_y else pyo.Reals, bounds=(-10, 10))

    m.o = pyo.Objective(expr=m.x + m.y, sense=pyo.maximize)

    m.c1 = pyo.Constraint(expr=-4 / 5 * m.x - 4 <= m.y)
    m.c2 = pyo.Constraint(expr=5 / 9 * m.x - 5 <= m.y)
    m.c3 = pyo.Constraint(expr=2 / 9 * m.x + 2 >= m.y)
    m.c4 = pyo.Constraint(expr=-1 / 2 * m.x + 3 >= m.y)

    # Continuous extreme points and bounds
    m.extreme_points = {
        (0.737704918, -4.590163934),
        (-5.869565217, 0.695652174),
        (1.384615385, 2.307692308),
        (7.578947368, -0.789473684),
    }

    m.continuous_bounds = pyo.ComponentMap()
    m.continuous_bounds[m.x] = (-5.869565217, 7.578947368)
    m.continuous_bounds[m.y] = (-4.590163934, 2.307692308)

    # Continuous extreme points and bounds for the case where an objective
    # constraint is added within a 100% relative gap of optimality or an
    # absolute gap of 6.789473684

    m.extreme_points_cut = {
        (45 / 14, -45 / 14),
        (-18 / 11, 18 / 11),
        (1.384615385, 2.307692308),
        (7.578947368, -0.789473684),
    }

    m.continuous_bounds_cut = pyo.ComponentMap()
    m.continuous_bounds_cut[m.x] = (-18 / 11, 7.578947368)
    m.continuous_bounds_cut[m.y] = (-45 / 14, 2.307692308)

    # Discrete feasible solutions and bounds
    feasible_sols = []
    x_lower_bound = None
    x_upper_bound = None
    y_lower_bound = None
    y_upper_bound = None

    x_lower = ceil(m.continuous_bounds[m.x][0])
    x_upper = floor(m.continuous_bounds[m.x][1])
    y_lower = ceil(m.continuous_bounds[m.y][0])
    y_upper = floor(m.continuous_bounds[m.y][1])
    cons = [m.c1, m.c2, m.c3, m.c4]
    for x_value in range(x_lower, x_upper + 1):
        for y_value in range(y_lower, y_upper + 1):
            m.x.set_value(x_value)
            m.y.set_value(y_value)
            is_feasible = True
            for con in cons:
                if not _is_satisfied(con):
                    is_feasible = False
                    break
            if is_feasible:
                if x_lower_bound is None or x_value < x_lower_bound:
                    x_lower_bound = x_value
                if x_upper_bound is None or x_value > x_upper_bound:
                    x_upper_bound = x_value
                if y_lower_bound is None or y_value < y_lower_bound:
                    y_lower_bound = y_value
                if y_upper_bound is None or y_value > y_upper_bound:
                    y_upper_bound = y_value
                feasible_sols.append(((x_value, y_value), x_value + y_value))
    m.discrete_feasible = sorted(feasible_sols, key=lambda sol: sol[1], reverse=True)
    m.discrete_bounds = pyo.ComponentMap()
    m.discrete_bounds[m.x] = (x_lower_bound, x_upper_bound)
    m.discrete_bounds[m.y] = (y_lower_bound, y_upper_bound)

    return m


def get_3d_polyhedron_problem():
    """
    Simple 3d polyhedron that is expressed using all types of linear constraints
    """
    m = pyo.ConcreteModel()
    m.x = pyo.Var([0, 1, 2], within=pyo.Reals)
    m.x[0].setlb(-1)
    m.x[0].setub(1)
    m.x[1].setlb(-2)
    m.x[1].setub(2)
    m.x[2].setlb(1)
    m.x[2].setub(2)

    def _constraint_switch_rule(m, i):
        if i == 0:
            return m.x[0] + m.x[1] <= 2
        elif i == 1:
            return -m.x[0] + m.x[1] <= 2
        elif i == 2:
            return m.x[0] + m.x[1] >= -2
        elif i == 3:
            return -m.x[0] + m.x[1] >= -2
        elif i == 4:
            return m.x[0] + m.x[1] + m.x[2] == 4

    m.c = pyo.Constraint([i for i in range(5)], rule=_constraint_switch_rule)

    m.o = pyo.Objective(expr=m.x[0] + m.x[2], sense=pyo.maximize)
    return m


def get_2d_unbounded_problem():
    """
    Simple 2d problem where the feasible region is unbounded, but the problem
    has an optimal solution.
    """
    m = pyo.ConcreteModel()
    m.x = pyo.Var(within=pyo.Reals)
    m.y = pyo.Var(within=pyo.Reals)

    m.o = pyo.Objective(expr=m.y - m.x)

    m.c1 = pyo.Constraint(expr=m.x <= 4)
    m.c2 = pyo.Constraint(expr=m.y >= 2)

    m.extreme_points = {(4, 2)}

    m.continuous_bounds = pyo.ComponentMap()
    m.continuous_bounds[m.x] = (float("-inf"), 4)
    m.continuous_bounds[m.y] = (2, float("inf"))
    return m


def get_2d_degenerate_lp():
    """
    Simple 2d problem that includes a redundant constraint such that three
    constraints are active at optimality.
    """
    m = pyo.ConcreteModel()

    m.x = pyo.Var(within=pyo.Reals, bounds=(-1, 3))
    m.y = pyo.Var(within=pyo.Reals, bounds=(-3, 2))

    m.obj = pyo.Objective(expr=m.x + 2 * m.y, sense=pyo.maximize)

    m.con1 = pyo.Constraint(expr=m.x + m.y <= 3)
    m.con2 = pyo.Constraint(expr=m.x + 2 * m.y <= 5)
    m.con3 = pyo.Constraint(expr=m.x + m.y >= -1)

    return m


def get_triangle_ip():
    """
    Simple 2d discrete problem where the feasible region looks like a 90-45-45
    right triangle and the optimal solutions fall along the hypotenuse, where
    x + y == 5.  Alternative near-optimal have integer objective values from 0 to 4.
    """
    var_max = 5
    m = pyo.ConcreteModel()
    m.x = pyo.Var(within=pyo.NonNegativeIntegers, bounds=(0, var_max))
    m.y = pyo.Var(within=pyo.NonNegativeIntegers, bounds=(0, var_max))

    m.o = pyo.Objective(expr=m.x + m.y, sense=pyo.maximize)
    m.c = pyo.Constraint(expr=m.x + m.y <= var_max)

    #
    # Enumerate all feasible solutions
    #
    feasible_sols = []
    for i in range(var_max + 1):
        for j in range(var_max + 1):
            if i + j <= var_max:
                feasible_sols.append(((i, j), i + j))
    feasible_sols = sorted(feasible_sols, key=lambda sol: sol[1], reverse=True)
    m.feasible_sols = feasible_sols
    #
    # Count of solutions from best to worst
    #
    m.num_ranked_solns = [6, 5, 4, 3, 2, 1]

    return m


def get_implied_bound_ip():
    """
    2d discrete problem where the bounds of z are impled by x and y. This
    facilitate testing cases where the impled bounds are tighter than the
    given bounds for the variable.
    """
    m = pyo.ConcreteModel()
    m.x = pyo.Var(within=pyo.NonNegativeIntegers, bounds=(0, 5))
    m.y = pyo.Var(within=pyo.NonNegativeIntegers, bounds=(0, 5))
    m.z = pyo.Var(within=pyo.NonNegativeIntegers, bounds=(0, 5))

    m.o = pyo.Objective(expr=m.x + m.z)

    m.c1 = pyo.Constraint(expr=m.x + m.y == 3)
    m.c2 = pyo.Constraint(expr=m.x + m.y + m.z <= 5)
    m.c3 = pyo.Constraint(expr=m.x + m.y + m.z >= 4)

    m.var_bounds = pyo.ComponentMap()
    m.var_bounds[m.x] = (0, 3)
    m.var_bounds[m.y] = (0, 3)
    m.var_bounds[m.z] = (1, 2)

    return m


def get_aos_test_knapsack(
    var_max, weights, values, capacity=None, capacity_fraction=1.0
):
    """
    Creates a knapsack problem, given arrays of weights and values, and
    returns all feasible solutions. The capacity represents the percent of the
    total max weight that can be selected (sum weights * var_max). The var_max
    parameter sets the upper bound on all variables, the max number of times
    they can be selected.
    """
    assert len(weights) == len(values), "weights and values must be the same length."
    assert (
        0 <= capacity_fraction and capacity_fraction <= 1
    ), "capacity_fraction must be between 0 and 1."

    num_vars = len(weights)
    if capacity is None:
        capacity = sum(weights) * var_max * capacity_fraction

    m = pyo.ConcreteModel()
    m.i = pyo.RangeSet(0, num_vars - 1)

    if var_max == 1:
        m.x = pyo.Var(m.i, within=pyo.Binary)
    else:
        m.x = pyo.Var(m.i, within=pyo.NonNegativeIntegers, bounds=(0, var_max))

    m.o = pyo.Objective(expr=sum(values[i] * m.x[i] for i in m.i), sense=pyo.maximize)

    m.c = pyo.Constraint(expr=sum(weights[i] * m.x[i] for i in m.i) <= capacity)

    var_domain = range(var_max + 1)
    all_combos = product(var_domain, repeat=num_vars)

    feasible_sols = []
    for sol in all_combos:
        if np.dot(sol, weights) <= capacity:
            feasible_sols.append((sol, np.dot(sol, values)))
    feasible_sols = sorted(feasible_sols, key=lambda sol: sol[1], reverse=True)
    m.ranked_solution_values = list(sorted([v for x, v in feasible_sols], reverse=True))
    m.num_ranked_solns = list(Counter([v for x, v in feasible_sols]).values())
    return m


def get_pentagonal_lp():
    """
    Pentagonal LP
    """
    var_max = 5
    m = pyo.ConcreteModel()
    m.x = pyo.Var(within=pyo.Reals, bounds=(0, 2 * var_max))
    m.y = pyo.Var(within=pyo.Reals, bounds=(0, 2 * var_max))
    m.z = pyo.Var(within=pyo.NonNegativeReals, bounds=(0, 2 * var_max))
    m.o = pyo.Objective(expr=m.z, sense=pyo.minimize)

    base_points = np.array(
        [
            [var_max, 2 * var_max, 0],
            [2 * var_max, var_max, 0],
            [3.0 * var_max / 2.0, 0, 0],
            [var_max / 2.0, 0, 0],
            [0, var_max, 0],
        ]
    )
    apex_point = np.array([var_max, var_max, var_max])

    m.c = pyo.ConstraintList()
    for i in range(5):
        vec_1 = base_points[i] - apex_point
        vec_2 = base_points[(i + 1) % var_max] - base_points[i]
        n = np.cross(vec_1, vec_2)
        m.c.add(
            n[0] * (m.x - apex_point[0])
            + n[1] * (m.y - apex_point[1])
            + n[2] * (m.z - apex_point[2])
            >= 0
        )

    return m


def get_pentagonal_pyramid_mip():
    """
    Pentagonal pyramid with integer coordinates in the first two dimensions and
    a third continuous dimension.
    """
    var_max = 5
    m = pyo.ConcreteModel()
    m.x = pyo.Var(within=pyo.Integers, bounds=(-var_max, var_max))
    m.y = pyo.Var(within=pyo.Integers, bounds=(-var_max, var_max))
    m.z = pyo.Var(within=pyo.NonNegativeReals, bounds=(0, var_max))
    m.o = pyo.Objective(expr=m.z, sense=pyo.maximize)

    base_points = np.array(
        [
            [0, var_max, 0],
            [var_max, 0, 0],
            [var_max / 2.0, -var_max, 0],
            [-var_max / 2.0, -var_max, 0],
            [-var_max, 0, 0],
        ]
    )
    apex_point = np.array([0, 0, var_max])

    m.c = pyo.ConstraintList()
    for i in range(5):
        vec_1 = base_points[i] - apex_point
        vec_2 = base_points[(i + 1) % var_max] - base_points[i]
        n = np.cross(vec_1, vec_2)
        m.c.add(
            n[0] * (m.x - apex_point[0])
            + n[1] * (m.y - apex_point[1])
            + n[2] * (m.z - apex_point[2])
            >= 0
        )
    #
    # Count of solutions from best to worst
    #
    m.num_ranked_solns = [1, 4, 2, 8, 2, 12, 4, 16, 4, 20]
    return m


def get_indexed_pentagonal_pyramid_mip():
    """
    Pentagonal pyramid with integer coordinates in the first two dimensions and
    a third continuous dimension.
    """
    var_max = 5
    m = pyo.ConcreteModel()
    m.x = pyo.Var([1, 2], within=pyo.Integers, bounds=(-var_max, var_max))
    m.z = pyo.Var(within=pyo.NonNegativeReals, bounds=(0, var_max))
    m.o = pyo.Objective(expr=m.z, sense=pyo.maximize)
    base_points = np.array(
        [
            [0, var_max, 0],
            [var_max, 0, 0],
            [var_max / 2.0, -var_max, 0],
            [-var_max / 2.0, -var_max, 0],
            [-var_max, 0, 0],
        ]
    )
    apex_point = np.array([0, 0, var_max])

    def _con_rule(m, i):
        vec_1 = base_points[i] - apex_point
        vec_2 = base_points[(i + 1) % var_max] - base_points[i]
        n = np.cross(vec_1, vec_2)
        expr = (
            n[0] * (m.x[1] - apex_point[0])
            + n[1] * (m.x[2] - apex_point[1])
            + n[2] * (m.z - apex_point[2])
        )
        return expr >= 0

    m.c = pyo.Constraint([i for i in range(5)], rule=_con_rule)
    m.num_ranked_solns = [1, 4, 2, 8, 2, 12, 4, 16, 4, 20]
    return m


def get_bloated_pentagonal_pyramid_mip():
    """
    Pentagonal pyramid with integer coordinates in the first two dimensions and
    a third continuous dimension. Bounds are artificially widened for obbt testing purposes
    """
    var_max = 5
    m = pyo.ConcreteModel()
    m.x = pyo.Var(within=pyo.Integers, bounds=(-2 * var_max, 2 * var_max))
    m.y = pyo.Var(within=pyo.Integers, bounds=(-2 * var_max, var_max))
    m.z = pyo.Var(within=pyo.NonNegativeReals, bounds=(0, 2 * var_max))
    m.var_bounds = pyo.ComponentMap()
    m.o = pyo.Objective(expr=m.z, sense=pyo.maximize)
    base_points = np.array(
        [
            [0, var_max, 0],
            [var_max, 0, 0],
            [var_max / 2.0, -var_max, 0],
            [-var_max / 2.0, -var_max, 0],
            [-var_max, 0, 0],
        ]
    )
    apex_point = np.array([0, 0, var_max])

    m.c = pyo.ConstraintList()
    for i in range(5):
        vec_1 = base_points[i] - apex_point
        vec_2 = base_points[(i + 1) % var_max] - base_points[i]
        n = np.cross(vec_1, vec_2)
        m.c.add(
            n[0] * (m.x - apex_point[0])
            + n[1] * (m.y - apex_point[1])
            + n[2] * (m.z - apex_point[2])
            >= 0
        )
    return m
